/*
FILE:     ConnectFormat.C
*/
/*
VERSION:  10.000
*/
/*
DATE:     6/6/2017
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2017 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/

/*!
** \file ConnectFormat.C
**
** \brief Implementation file for ConnectFormat class.
*/

#include <stdlib.h>
#include <math.h>

#include "ConnectFormat.h"
#include "ConnectFormat_global.h"
#include "Exceptions.h"
#include "utillib.h"


void ConnectFormat::clear()
{
       _has_chiral_center = false;
       _nheavyatoms = 0;
       _chem_comp_data.clear();
       _atoms.clear();
       _order_atoms.clear();
       _bonds.clear();
       _atomOrder.clear();
}

ConnectFormat& ConnectFormat::operator=(const ConnectFormat& drug)
{
       if (this != &drug) {
            clear();
            _has_chiral_center = drug._has_chiral_center;
            _nheavyatoms = drug._nheavyatoms;
            _chem_comp_data = drug._chem_comp_data;
            _atoms = drug._atoms;
            _bonds = drug._bonds;
            _atomOrder = drug._atomOrder;
       }
       return (*this);
}

const bool& ConnectFormat::has_chiral_center() const
{
       return _has_chiral_center;
}

const int& ConnectFormat::nheavyatoms() const
{
       return _nheavyatoms;
}

const int ConnectFormat::natoms() const
{
       return ((int) _atoms.size());
}

const int ConnectFormat::nbonds() const
{
       return ((int) _bonds.size());
}

const std::string ConnectFormat::drugname() const
{
       return getMetaData("id");
}

const std::string ConnectFormat::formula() const
{
       return getMetaData("formula");
}

const std::string ConnectFormat::chemical_name() const
{
       return getMetaData("name");
}

const std::string ConnectFormat::synonym() const
{
       return getMetaData("pdbx_synonyms");
}

const std::string ConnectFormat::getMetaData(const std::string& item) const
{
       std::map<std::string, std::string>::const_iterator
           mpos = _chem_comp_data.find(item);
       if (mpos != _chem_comp_data.end()) return mpos->second;
       return "";
}

const AtomFormat& ConnectFormat::find_atom(const std::string& name) const
{
       std::map<std::string, unsigned int>::const_iterator
           pos = _atomOrder.find(name);
       if (pos != _atomOrder.end())
            return _atoms[pos->second];
       else throw NotFoundException();
}

const AtomFormat& ConnectFormat::find_atom(const int& pos) const
{
       if (pos >= 0 && pos < (int) _atoms.size())
            return _atoms[pos];
       else throw out_of_range("");
}

const int ConnectFormat::find_atom_pos(const std::string& name) const
{
       std::map<std::string, unsigned int>::const_iterator
           pos = _atomOrder.find(name);
       if (pos != _atomOrder.end())
            return (int) pos->second;
       else return -1;
}

const std::vector<AtomFormat>& ConnectFormat::atoms() const
{
       return _atoms;
}

const std::vector<AtomFormat>& ConnectFormat::order_atoms() const
{
       if (!_order_atoms.empty())
            return _order_atoms;
       else return _atoms;
}

const std::vector<std::vector<std::string> >& ConnectFormat::bonds() const
{
       return _bonds;
}

const std::vector<std::vector<std::string> > ConnectFormat::atomArray() const
{
       std::vector<std::vector<std::string> > atom_vector;
       atom_vector.clear();

       for (std::vector<AtomFormat>::const_iterator
            pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            atom_vector.push_back(pos->atom());
       }
       return atom_vector;
}

const std::vector<std::pair<std::string, std::string> > ConnectFormat::getAtomList() const
{
       std::vector<std::pair<std::string, std::string> > pair_list;
       pair_list.clear();

       for (std::vector<AtomFormat>::const_iterator
            pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            pair_list.push_back(std::make_pair(pos->atomtype(), pos->atomname()));
       }
       return pair_list;
}

const std::vector<std::pair<std::string, std::string> > ConnectFormat::getBondList() const
{
       std::vector<std::pair<std::string, std::string> > pair_list;
       pair_list.clear();

       for (std::vector<std::vector<std::string> >::const_iterator
            pos = _bonds.begin(); pos != _bonds.end(); ++pos) {
            pair_list.push_back(std::make_pair((*pos)[0], (*pos)[1]));
       }
       return pair_list;
}

const std::map<std::string, std::vector<std::string> > ConnectFormat::getLinkedAtoms() const
{
       std::map<std::string, std::vector<std::string> > linked_atom_mapping;
       linked_atom_mapping.clear();

       std::vector<std::string> data;
       for (std::vector<AtomFormat>::const_iterator
            pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            data.clear();
            const std::vector<std::vector<std::string> >& bonds = pos->bonds();
            for (std::vector<std::vector<std::string> >::const_iterator
                 vpos = bonds.begin(); vpos != bonds.end(); ++vpos) {
                 data.push_back((*vpos)[0]);
            }
            linked_atom_mapping.insert(std::make_pair(pos->atomname(), data));
       }
       return linked_atom_mapping;
}

const std::set<std::string> ConnectFormat::getLeavingdAtoms() const
{
       std::set<std::string> leaving_atom_set;
       leaving_atom_set.clear();

       for (std::vector<AtomFormat>::const_iterator
            pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (pos->leaving_atom_flag() != "Y") continue;
            leaving_atom_set.insert(pos->atomname());
       }
       return leaving_atom_set;
}

void ConnectFormat::getChiralCenterAtoms(const std::string& center_atom_name, std::vector<std::string>& atomnames,
                                         std::vector<std::vector<double> >& coords) const
{
       atomnames.clear();
       coords.clear();

       std::map<std::string, unsigned int>::const_iterator
           pos = _atomOrder.find(center_atom_name);
       if (pos == _atomOrder.end()) return;

       std::vector<AtomFormat> Atoms;
       Atoms.clear();
       Atoms.push_back(_atoms[pos->second]);

       std::vector<std::vector<std::string> > bonds = Atoms[0].bonds();
       for (std::vector<std::vector<std::string> >::const_iterator
            vpos = bonds.begin(); vpos != bonds.end(); ++vpos) {
            pos = _atomOrder.find((*vpos)[0]);
            if (pos == _atomOrder.end()) continue;
            if (_atoms[pos->second].atomtype() == "H" ||
                _atoms[pos->second].atomtype() == "D") continue;
            Atoms.push_back(_atoms[pos->second]);
       }
       if (Atoms.size() < 4) return;

       bool has_ideal_coor = true;
       bool has_model_coor = true;
       for (std::vector<AtomFormat>::const_iterator
            vpos = Atoms.begin(); vpos != Atoms.end(); ++vpos) {
            if (vpos->get_value(8).empty()) has_model_coor = false;
            if (vpos->get_value(9).empty()) has_model_coor = false;
            if (vpos->get_value(10).empty()) has_model_coor = false;
            if (vpos->get_value(11).empty()) has_ideal_coor = false;
            if (vpos->get_value(12).empty()) has_ideal_coor = false;
            if (vpos->get_value(13).empty()) has_ideal_coor = false;
       }
       if (!has_ideal_coor && !has_model_coor) return;

       std::vector<double> coord;
       for (std::vector<AtomFormat>::const_iterator
            vpos = Atoms.begin(); vpos != Atoms.end(); ++vpos) {
            atomnames.push_back(vpos->atomname());
            coord.clear();
/*
            if (has_ideal_coor) {
                 coord.push_back(atof(vpos->get_value(11).c_str()));
                 coord.push_back(atof(vpos->get_value(12).c_str()));
                 coord.push_back(atof(vpos->get_value(13).c_str()));
            } else {
                 coord.push_back(atof(vpos->get_value(8).c_str()));
                 coord.push_back(atof(vpos->get_value(9).c_str()));
                 coord.push_back(atof(vpos->get_value(10).c_str()));
            }
*/
            if (has_model_coor) {
                 coord.push_back(atof(vpos->get_value(8).c_str()));
                 coord.push_back(atof(vpos->get_value(9).c_str()));
                 coord.push_back(atof(vpos->get_value(10).c_str()));
            } else {
                 coord.push_back(atof(vpos->get_value(11).c_str()));
                 coord.push_back(atof(vpos->get_value(12).c_str()));
                 coord.push_back(atof(vpos->get_value(13).c_str()));
            }
            coords.push_back(coord);
       }
}

void ConnectFormat::getTerminalAtoms(std::vector<std::string>& atomnames) const
{
       atomnames.clear();

       // map key: atom type
       // value: terminal atom list
       std::map<std::string, std::vector<std::string> > terminals;
       _getTerminalAtoms(terminals);

       for (std::map<std::string, std::vector<std::string> >::const_iterator
            mpos = terminals.begin(); mpos != terminals.end(); ++mpos) {
            for (std::vector<std::string>::const_iterator
                 pos = mpos->second.begin(); pos != mpos->second.end(); ++pos) {
                 atomnames.push_back(*pos);
            }
       }
}

std::string ConnectFormat::getTerminalAtom(const std::string& type) const
{
       // map key: atom type
       // value: terminal atom list
       std::map<std::string, std::vector<std::string> > terminals;
       _getTerminalAtoms(terminals);
       std::map<std::string, std::vector<std::string> >::const_iterator
           mpos = terminals.find(type);
       if (mpos != terminals.end()) return mpos->second[0];
       return "";
}

bool ConnectFormat::isTerminalAtom(const std::string& atomName) const
{
       std::map<std::string, unsigned int>::const_iterator
           pos = _atomOrder.find(atomName);
       if (pos != _atomOrder.end()) {
            if (String::IsEqual(_atoms[pos->second].leaving_atom_flag(), "Y",
                                Char::eCASE_INSENSITIVE)) return true;
       } else return true;
       return false;
}

void ConnectFormat::getInterSubComponentBonds(std::vector<std::vector<std::string> >& links) const
{
       links.clear();

       std::vector<std::string> data;

       for (std::vector<std::vector<std::string> >::const_iterator
            bpos = _bonds.begin(); bpos != _bonds.end(); ++bpos) {
            std::map<std::string, unsigned int>::const_iterator
                mpos1 = _atomOrder.find((*bpos)[0]);
            if (mpos1 == _atomOrder.end()) continue;
            std::map<std::string, unsigned int>::const_iterator
                mpos2 = _atomOrder.find((*bpos)[1]);
            if (mpos2 == _atomOrder.end()) continue;
            if (_atoms[mpos1->second].get_value(14) == _atoms[mpos2->second].get_value(14) &&
                _atoms[mpos1->second].get_value(15) == _atoms[mpos2->second].get_value(15))
                 continue;
            // vector[0]: comp_id_1
            // vector[1]: numbering_1
            // vector[2]: atom_id_1
            // vector[3]: component_id_1
            // vector[4]: comp_id_2
            // vector[5]: numbering_2
            // vector[6]: atom_id_2
            // vector[7]: component_id_2
            // vector[8]: bond_type
            // vector[9]: polyper_type_1
            // vector[10]: polyper_type_2
            data.clear();
            data.push_back(_atoms[mpos1->second].get_value(14));
            data.push_back(_atoms[mpos1->second].get_value(15));
            data.push_back(_atoms[mpos1->second].get_value(16));
            data.push_back(_atoms[mpos1->second].get_value(19));
            data.push_back(_atoms[mpos2->second].get_value(14));
            data.push_back(_atoms[mpos2->second].get_value(15));
            data.push_back(_atoms[mpos2->second].get_value(16));
            data.push_back(_atoms[mpos2->second].get_value(19));
            data.push_back((*bpos)[2]);
            data.push_back(_atoms[mpos1->second].get_value(17));
            data.push_back(_atoms[mpos2->second].get_value(17));
            links.push_back(data);
       }
}

bool ConnectFormat::Read(Block &block)
{
       ISTable *Table = getTablePtr(block, "chem_comp");
       if (!Table) return false;

       std::string cs;
       for (int i = 0; i < NUM_CHEM_COMP; ++i) {
            get_value_clean(cs, Table, 0, _chem_comp[i]);
            if (cs.empty()) continue;
            _chem_comp_data.insert(std::make_pair(_chem_comp[i], cs));
       }
       if (_chem_comp_data.find("pdbx_type") == _chem_comp_data.end()) {
            _chem_comp_data.insert(std::make_pair("pdbx_type", "HETAIN"));
       }

       Table = getTablePtr(block, "chem_comp_atom");
       if (!Table) return false;

       std::vector<std::string> data;
       AtomFormat atom;
       int rowNo = Table->GetNumRows();
       for (int i = 0; i < rowNo; i++) {
            data.clear();
            for (int j = 0; j < NUM_CHEM_COMP_ATOM; ++j) {
                 get_value_clean(cs, Table, i, _chem_comp_atom[j]);
                 data.push_back(cs);
            }
            if (data[0].empty()) continue;

            if (!data[2].empty() && data[2] != "H" && data[2] != "D") _nheavyatoms++;
            if (!data[7].empty() && data[7] != "N") _has_chiral_center = true;

            atom.clear();
            atom.set_atom(data);
            _atomOrder.insert(std::make_pair(data[0], _atoms.size()));
            _atoms.push_back(atom);
       }

       Table = getTablePtr(block, "chem_comp_bond");
       if (Table) {
            std::vector<std::string> bond;
            rowNo = Table->GetNumRows();
            for (int i = 0; i < rowNo; i++) {
                 data.clear();
                 for (int j = 0; j < NUM_CHEM_COMP_BOND; ++j) {
                      get_value_clean(cs, Table, i, _chem_comp_bond[j]);
                      data.push_back(cs);
                 }
                 _bonds.push_back(data);

                 bond.clear();
                 for (unsigned int j = 1; j < data.size(); ++j) {
                      bond.push_back(data[j]);
                 }
                 std::map<std::string, unsigned int>::const_iterator
                     mpos = _atomOrder.find(data[0]);
                 if (mpos != _atomOrder.end()) {
                      _atoms[mpos->second].add_bond(bond);
                 }
                 mpos = _atomOrder.find(data[1]);
                 if (mpos != _atomOrder.end()) {
                      bond[0] = data[0];
                      _atoms[mpos->second].add_bond(bond);
                 }
            }
       }

       _getOrderedAatoms();

       return true;
}

void ConnectFormat::_getOrderedAatoms()
{
       if (!_findSubComponentBreak()) return;

       std::map<int, int> index;
       index.clear();
       bool has_polymer = false;
       bool has_non_polymer = false;
       for (unsigned int i = 0; i < _atoms.size(); ++i) {
            int j = atoi(_atoms[i].get_value(15).c_str()) * _atoms.size() + i;
            index.insert(std::make_pair(j, i));
            if (_atoms[i].get_value(17) == "polymer") has_polymer = true;
            if (_atoms[i].get_value(17) == "non-polymer") has_non_polymer = true;
       }

       _order_atoms.clear();
       for (std::map<int, int>::iterator
            pos = index.begin(); pos != index.end(); ++pos) {
            _order_atoms.push_back(_atoms[pos->second]);
       }
       if (!has_polymer || !has_non_polymer) return;

       std::vector<AtomFormat> atoms;
       atoms.clear();
       for (std::vector<AtomFormat>::const_iterator
            pos = _order_atoms.begin(); pos != _order_atoms.end(); ++pos) {
            if (pos->get_value(17) == "non-polymer") continue;
            atoms.push_back(*pos);
       }
       for (std::vector<AtomFormat>::const_iterator
            pos = _order_atoms.begin(); pos != _order_atoms.end(); ++pos) {
            if (pos->get_value(17) == "polymer") continue;
            atoms.push_back(*pos);
       }
       _order_atoms = atoms;
}

bool ConnectFormat::_findSubComponentBreak()
{
       std::string numbering = "";
       int count = 0;
       bool is_hydrogen_break = false;
       std::set<unsigned int> residue_breaks;
       for (unsigned int i = 0; i < _atoms.size(); ++i) {
            if (_atoms[i].get_value(15) != numbering) {
                 numbering = _atoms[i].get_value(15);
                 count++;
            }
            // insert sub component break point if two sub components
            // have different residue names
            if (i && _atoms[i].get_value(14) != _atoms[i-1].get_value(14))
                 residue_breaks.insert(i);
            // insert sub component break point between heavy atom & hydrogen
            if (i && (_atoms[i].get_value(2) == "H" || _atoms[i].get_value(2) == "D") &&
                _atoms[i-1].get_value(2) != "H" && _atoms[i-1].get_value(2) != "D") {
                 residue_breaks.insert(i);
                 is_hydrogen_break = true;
            }
       }
       if (count > 0) return true;

       if (is_hydrogen_break && residue_breaks.size() == 1) residue_breaks.clear();

       // convert residue point break set into sub component range pair
       std::vector<std::pair<unsigned int, unsigned int> > pair_array;
       pair_array.clear();
       if (residue_breaks.empty()) // whole ligand has single range pair
            pair_array.push_back(std::make_pair(0, _atoms.size() - 1));
       else {
            std::vector<unsigned int> break_array;
            break_array.clear();
            for (std::set<unsigned int>::iterator
                 pos = residue_breaks.begin(); pos != residue_breaks.end(); ++pos) {
                 break_array.push_back(*pos);
            }

            // insert first sub component range pair
            pair_array.push_back(std::make_pair(0, break_array[0] - 1));
            // insert middle sub component(s) range pair
            for (unsigned int i = 0; i < break_array.size() - 1; ++i) {
                 pair_array.push_back(std::make_pair(break_array[i],
                                                     break_array[i+1]-1));
            }
            // insert last sub component range pair
            unsigned int i = break_array.size() - 1;
            pair_array.push_back(std::make_pair(break_array[i], _atoms.size() - 1));
       }

       for (std::vector<std::pair<unsigned int, unsigned int> >::iterator
            pos = pair_array.begin(); pos != pair_array.end(); ++pos) {
            _getSameSubComponentBreak(pos->first, pos->second, residue_breaks);
       }

       std::string type = "non-polymer";
       if (!residue_breaks.empty() && getMetaData("pdbx_release_status") == "REF_ONLY")
            type = "polymer";
       count = 1;
       bool first_hydrogen = true;
       for (unsigned int i = 0; i < _atoms.size(); ++i) {
            if (residue_breaks.find(i) != residue_breaks.end()) {
                 count++;
                 if (first_hydrogen && (_atoms[i].get_value(2) == "H" || _atoms[i].get_value(2) == "D")) {
                      count = 1;
                      first_hydrogen = false;
                 }
            }
            if (_atoms[i].get_value(15).empty())
                 _atoms[i].set_value(15, String::IntToString(count));
            if (_atoms[i].get_value(17).empty()) _atoms[i].set_value(17, type);
       }

       if (!residue_breaks.empty())
            return true;
       else return false;
}

void ConnectFormat::_getSameSubComponentBreak(const unsigned int& begin, const unsigned int& end,
                                              std::set<unsigned int>& residue_breaks)
{
       for (unsigned int i = begin + 1; i <= end; ++i) {
            if (_atoms[i].get_value(16) == _atoms[begin].get_value(16)) residue_breaks.insert(i);
       }
}

void ConnectFormat::_getTerminalAtoms(std::map<std::string, std::vector<std::string> >&
                                      terminals) const
{
       terminals.clear();

       std::string id = "";
       std::map<std::string, std::string>::const_iterator
           mmpos = _chem_comp_data.find("id");
       if (mmpos != _chem_comp_data.end()) id = mmpos->second;

       std::vector<std::string> data;
       if (id == "ACE") {
            data.clear();
            data.push_back("C");
            terminals.insert(std::make_pair("C", data));
            return;
       } else if (id == "NH2") {
            data.clear();
            data.push_back("N");
            terminals.insert(std::make_pair("N", data));
            return;
       }

       if (_atoms.empty()) return;


       if (_atoms.size() == 1) {
            data.clear();
            data.push_back(_atoms[0].atomname());
            terminals.insert(std::make_pair(_atoms[0].atomtype(), data));
            return;
       }

       for (std::vector<AtomFormat>::const_iterator
            pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (!String::IsEqual(pos->leaving_atom_flag(), "Y",
                                 Char::eCASE_INSENSITIVE)) continue;

            const std::vector<std::vector<std::string> >& bonds = pos->bonds();
            for (std::vector<std::vector<std::string> >::const_iterator
                 bpos = bonds.begin(); bpos != bonds.end(); ++bpos) {
                 std::map<std::string, unsigned int>::const_iterator
                     mpos = _atomOrder.find((*bpos)[0]);
                 if (mpos == _atomOrder.end()) continue;
                 if (!String::IsEqual(_atoms[mpos->second].leaving_atom_flag(), "Y",
                                      Char::eCASE_INSENSITIVE)) continue;

                 std::map<std::string, std::vector<std::string> >::iterator
                     tpos = terminals.find(_atoms[mpos->second].atomtype());
                 if (tpos != terminals.end())
                      tpos->second.push_back(_atoms[mpos->second].atomname());
                 else {
                      data.clear();
                      data.push_back(_atoms[mpos->second].atomname());
                      terminals.insert(std::make_pair(_atoms[mpos->second].atomtype(),
                                       data));
                 }
            }
       }
}
