/*
FILE:     Residue.C
*/
/*
VERSION:  10.000
*/
/*
DATE:     6/6/2017
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2017 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>

#include <math.h>

#include "BondUtil.h"
#include "CompositeIndex.h"
#include "Element.h"
#include "GraphMatch.h"
#include "MatrixUtil.h"
#include "Residue.h"
#include "Residue_global.h"
#include "SeqCodeUtil.h"
#include "TypeDef.h"
#include "utillib.h"

IntegerStringOrder::IntegerStringOrder()
{
       _int_value = 0;
       _str_value.clear();
}

const int IntegerStringOrder::int_value() const { return _int_value; }
const std::string &IntegerStringOrder::str_value() const { return _str_value; }
void IntegerStringOrder::set_int_value(const int& a)   { _int_value = a; }
void IntegerStringOrder::set_str_value(const std::string &a) { _str_value = a; }

IntegerStringOrder& IntegerStringOrder::operator=(const IntegerStringOrder &index)
{
       _int_value = index._int_value;
       _str_value = index._str_value;
       return (*this);
}

bool IntegerStringOrder::operator()(const IntegerStringOrder &index1,
                                    const IntegerStringOrder &index2) const
{
       return (index1 < index2);
}

bool IntegerStringOrder::operator<(const IntegerStringOrder &index) const
{
       if (_int_value < index._int_value)
            return true;
       else if (_int_value == index._int_value) {
            if (_str_value < index._str_value)
                 return true;
            else return false;
       } else return false;
}

bool IntegerStringOrder::operator==(const IntegerStringOrder &index) const
{
       if (_int_value == index._int_value &&
           _str_value == index._str_value)
            return true;
       else return false;
}

Residue::Residue()
{
       _ccDic = NULL;
       _logIo = NULL;
       _messageIo = NULL;
       clear(); 
}

void Residue::clear()
{
       _token.clear();
       _ResName.clear();
       _OrigResName.clear();
       _chnid.clear();
       _res_no.clear();
       _pdb_chnid.clear();
       _pdb_res_no.clear();
       _ins_code.clear();
       _alt_loc.clear();
       _alt_loc_list.clear();
       _ter_flag = 0;
       _atoms.clear();
       _type = MONOMER_TYPE_UNK;
       _chain_type.clear();
       _entity_id.clear();
       _chain_break = 0;
       _wrong_connectivity = false;
       _index = 0;
       _tls_group = 0;
       _chain_index = 0;
       _position = 0;
       _uniqueAtomNames.clear();
       _atomNames.clear();
       _atomNameAlts.clear();
       _atomNameAlts_Orininal.clear();
       _atomOrder.clear();
       _missing.clear();
       _extras.clear();
       _atom_pos = _atomOrder.begin();
}

void Residue::_reset()
{
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (*pos) delete *pos;
       }
       clear();
}

void Residue::_UpdateIndices()
{
       _uniqueAtomNames.clear();
       _atomNames.clear();
       _atomNameAlts.clear();
       int i = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            _uniqueAtomNames.insert((*pos)->getAtomIndex());
            _atomNames.insert(std::make_pair((*pos)->getAtomIndex(), i));
            _atomNameAlts.insert(std::make_pair((*pos)->getAtomAltIndex(), i));
            i++;
       }
}

void Residue::_writeAtomRecord(Atom *atom, std::string& record)
{
       record.clear();

       std::string chn_id = atom->pdb_chnid();
       if (chn_id.empty()) chn_id = " ";
       std::string inscode = atom->ins_code();
       if (inscode.empty()) inscode = " ";
       std::string alt_loc = atom->alt_loc();
       if (alt_loc.empty()) alt_loc = " "; 

       char buffer[100];

       sprintf(buffer, FORMAT, atom->atnum().c_str(), atom->pdb_atmnam().c_str(),
            alt_loc[0], atom->pdb_resnam().c_str(), chn_id.c_str(),
            atom->pdb_resnum().c_str(), inscode[0], atom->x(), atom->y(),
            atom->z(), atof(atom->occ().c_str()), atof(atom->t_fct().c_str()));
       record += buffer;
}

void Residue::_writeErrorMessage(Atom *atom1, Atom *atom2, const std::string& type, const std::string& message, const int& Mol_ID)
{
       std::string cs, buffer;
       cs.clear();
       if (Mol_ID < 0)
            cs += "\n";
       else {
            cs += "\nIn model " + String::IntToString(Mol_ID);
       }
       cs += message;

       _writeAtomRecord(atom1, buffer); cs += buffer;
       _writeAtomRecord(atom2, buffer); cs += buffer;

       _messageIo->insertMessage(type, "label", "true");
       _messageIo->insertMessage(type + "_print", "error", cs);
}

void Residue::_writeOccupancyMessage(const std::string &name, const float& occupancy, const int& Mol_ID)
{
       _messageIo->insertMessage("occupancy", "label", "true");

       std::string error = "";
       if (Mol_ID >= 0) error += "In model " + String::IntToString(Mol_ID) + ",";

       error += "Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + name;
       error += "' has occupancy " + FloatToString(occupancy, 0, 2) + ".";
       _messageIo->insertMessage("occupancy_print", "warning", error);
}

void Residue::_writeOccupancyError(const std::string &name, const float& occupancy, const int& Mol_ID)
{
       std::string error = "";
       if (Mol_ID >= 0) error += "In model " + String::IntToString(Mol_ID) + ",";

       error += "Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + name;
       error += "' has occupancy " + FloatToString(occupancy, 0, 2) + ".\n";
       _logIo->message(error.c_str());
}

const std::string& Residue::token() const { return _token; }
const std::string& Residue::ResName() const { return _ResName; }
const std::string& Residue::OrigResName() const { return _OrigResName; }
const std::string& Residue::chnid() const { return _chnid; }
const std::string& Residue::res_no() const { return _res_no; }
const std::string& Residue::pdb_chnid() const { return _pdb_chnid; }
const std::string& Residue::pdb_res_no() const { return _pdb_res_no; }
const std::string& Residue::ins_code() const { return _ins_code; }
const std::string& Residue::alt_loc() const { return _alt_loc; }
const bool Residue::has_alt_loc() const { return _alt_loc_list.size() > 0; }
const int& Residue::ter_flag() const { return _ter_flag; }
const int& Residue::type() const { return _type; }
const std::string& Residue::chain_type() const { return _chain_type; }
const std::string& Residue::entity_id() const { return _entity_id; }
const int& Residue::chain_break() const { return _chain_break; }
const bool& Residue::wrong_connectivity() const { return _wrong_connectivity; }
const int& Residue::index() const { return _index; }
const int& Residue::tls_group() const { return _tls_group; }
const int& Residue::chain_index() const { return _chain_index; }
const int& Residue::position() const { return _position; }
const std::vector<std::string>& Residue::missing() const { return _missing; }
const std::vector<std::string>& Residue::extras() const { return _extras; }
const bool Residue::is_single_atom() const { return (_uniqueAtomNames.size() == 1); }
const unsigned int Residue::AtomNumbers() const { return _atomNames.size(); }
const unsigned int Residue::NumAtoms() const { return _atoms.size(); }

const int Residue::NonHydrogenAtomNumbers()
{
       std::set<std::string> index;
       index.clear();
       for (std::vector<Atom*>::const_iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->atom_type() == "H" || (*pos)->atom_type() == "D") continue;
            index.insert((*pos)->atmtype());
       }
       return (index.size());
}

const int Residue::ca_or_p_atom_only() const
{
       std::set<std::string> name_set;
       name_set.clear();
       for (std::vector<Atom*>::const_iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            name_set.insert((*pos)->pdb_atmnam());
            if (name_set.size() > 1) return 0;
       }

       if (name_set.size() == 1) {
            std::set<std::string>::const_iterator spos = name_set.begin();
            if (*spos == "CA") return CAAtom_ONLY;
            if (*spos == "P") return PAtom_ONLY;
       }

       return 0;
       
}

void Residue::setCCDic(ConnectDic *ccdic) { _ccDic = ccdic; }
void Residue::setLog(LogUtil *logPt) { _logIo = logPt; }
void Residue::setMessage(MessageUtil *message) { _messageIo = message; }
void Residue::UpdateIndices() { _UpdateIndices(); }
void Residue::set_token(const std::string &a) { _token = a; }
void Residue::set_ResName(const std::string &a) { _ResName = a; }
void Residue::set_chnid(const std::string &a) { _chnid = a; }
void Residue::set_res_no(const std::string& a) { _res_no = a; }
void Residue::set_pdb_chnid(const std::string& a) { _pdb_chnid = a; }
void Residue::set_pdb_res_no(const std::string& a) { _pdb_res_no = a; }
void Residue::set_ins_code(const std::string& a) { _ins_code = a; }
void Residue::set_type(const int& a) { _type = a; }
void Residue::set_chain_type(const std::string& a) { _chain_type = a; }
void Residue::set_entity_id(const std::string& a) { _entity_id = a; }
void Residue::set_chain_break(const int& a) { _chain_break = a; }
void Residue::set_index(const int& a) { _index = a; }
void Residue::set_tls_group(const int& a) { _tls_group = a; }
void Residue::set_chain_index(const int& a) { _chain_index = a; }
void Residue::set_position(const int& a) { _position = a; }

void Residue::set_alt_loc()
{
       _alt_loc.clear();
       _alt_loc_list.clear();

       std::set<std::string> t_set;
       t_set.clear();
       bool has_no_alt_loc = false;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->alt_loc().empty()) {
                 has_no_alt_loc = true;
                 continue;
            }
            if (t_set.find((*pos)->alt_loc()) != t_set.end()) continue;
            t_set.insert((*pos)->alt_loc());
            _alt_loc_list.push_back((*pos)->alt_loc());
       }
       if (has_no_alt_loc) _alt_loc_list.clear();
       if (_alt_loc_list.size() == 1) _alt_loc = _alt_loc_list[0];
}

Atom* Residue::atoms(const int& pos)
{
       if (pos >= 0 && pos < (int) _atoms.size())
            return _atoms[pos];
       else return NULL;
}

Atom* Residue::GetFirstAtom()
{
       _atom_pos = _atomOrder.begin();
       if (_atom_pos != _atomOrder.end()) {
            return _atoms[_atom_pos->second];
       } else return NULL;
}

Atom* Residue::GetNextAtom()
{
       if (_atom_pos == _atomOrder.end()) return NULL;

       ++_atom_pos;
       if (_atom_pos != _atomOrder.end())
            return _atoms[_atom_pos->second];
       else return NULL;
}

void Residue::GetAtomNameList(std::list<std::string>& name_list)
{
       name_list.clear();

       std::set<std::string> name_set;
       name_set.clear();
       for (std::multimap<IntegerStringOrder, int>::const_iterator pos = _atomOrder.begin(); pos != _atomOrder.end(); ++pos) {
            if (name_set.find(_atoms[pos->second]->pdb_atmnam()) != name_set.end()) continue;
            name_set.insert(_atoms[pos->second]->pdb_atmnam());
            name_list.push_back(_atoms[pos->second]->pdb_atmnam());
       }
}

void Residue::GetAllBonds(std::vector<Atom*>& f_bond, std::vector<Atom*>& e_bond)
{
       f_bond.clear();
       e_bond.clear();
       for (unsigned int i = 0; i < _atoms.size() - 1; i++) {
            for (unsigned int j = i + 1; j < _atoms.size(); j++) {
                 if ((_atoms[i]->alt_loc().empty() || _atoms[j]->alt_loc().empty() || _atoms[i]->alt_loc() == _atoms[j]->alt_loc()) &&
                      BondUtil::is_a_bond(_atoms[i], _atoms[j])) {
                      f_bond.push_back(_atoms[i]);
                      e_bond.push_back(_atoms[j]);
                 }
            }
       }
}

void Residue::insert_a_atom(Atom* atom, const int& Mol_ID)
{
       std::multimap<std::string, int>::iterator
           pos = _atomNameAlts.find(atom->getAtomAltIndex());
       if (pos != _atomNameAlts.end()) {
            _writeErrorMessage(atom, _atoms[pos->second], "alt_code", "Atoms with same alternate conformation ID:\n", Mol_ID);
       }
       _atomNameAlts.insert(std::make_pair(atom->getAtomAltIndex(), _atoms.size()));

       pos = _atomNames.find(atom->getAtomIndex());
       if (pos != _atomNames.end()) {
            if ((!atom->alt_loc().empty() && _atoms[pos->second]->alt_loc().empty()) || (atom->alt_loc().empty() && !_atoms[pos->second]->alt_loc().empty()))
                 _writeErrorMessage(atom, _atoms[pos->second], "alt_code", "Atoms with same alternate conformation ID:\n", Mol_ID);
       }
       _uniqueAtomNames.insert(atom->getAtomAltIndex());
       _atomNames.insert(std::make_pair(atom->getAtomIndex(), _atoms.size()));

       IntegerStringOrder idx;
       idx.set_int_value(_atoms.size());
       idx.set_str_value(atom->alt_loc());
       _atomOrder.insert(std::make_pair(idx, _atoms.size()));

       if (atom->ter_flag()) _ter_flag = atom->ter_flag();
       _atoms.push_back(atom);
}

void Residue::delete_a_atom(const std::string& name, std::set<long>& atom_set)
{
       std::string cs = CompositeIndex::getIndex(_ResName, name);
       std::pair<std::multimap<std::string, int>::iterator, std::multimap<std::string, int>::iterator> range = _atomNames.equal_range(cs);
       if (range.first == range.second) return;

       std::set<int> index_set;
       index_set.clear();
       for (std::multimap<std::string, int>::iterator pos = range.first; pos != range.second; ++pos) {
            index_set.insert(pos->second);
       }

       std::vector<Atom*> atoms;
       atoms.clear();
       int i = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (index_set.find(i) != index_set.end()) {
                 atom_set.insert((long) (*pos));
                 delete *pos;
            } else atoms.push_back(*pos);
            i++;
       }
       _atoms = atoms;

       _atomNameAlts_Orininal.clear();
       _UpdateIndices();

       IntegerStringOrder idx;
       _atomOrder.clear();
       i = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            idx.set_int_value(i);
            idx.set_str_value((*pos)->alt_loc());
            _atomOrder.insert(std::make_pair(idx, i));
            i++;
       }

       _atom_pos = _atomOrder.begin();
}

void Residue::remove_a_atom(const std::string& name, std::vector<Atom*>& atom_list)
{
       atom_list.clear();
       std::string cs = CompositeIndex::getIndex(_ResName, name);
       std::pair<std::multimap<std::string, int>::iterator, std::multimap<std::string, int>::iterator> range = _atomNames.equal_range(cs);
       if (range.first == range.second) return;

       std::set<int> index_set;
       index_set.clear();
       for (std::multimap<std::string, int>::iterator pos = range.first; pos != range.second; ++pos) {
            index_set.insert(pos->second);
       }

       std::vector<Atom*> atoms;
       atoms.clear();
       int i = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (index_set.find(i) != index_set.end())
                 atom_list.push_back(*pos);
            else atoms.push_back(*pos);
            i++;
       }
       _atoms = atoms;

       _atomNameAlts_Orininal.clear();
       _UpdateIndices();

       IntegerStringOrder idx;
       _atomOrder.clear();
       i = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            idx.set_int_value(i);
            idx.set_str_value((*pos)->alt_loc());
            _atomOrder.insert(std::make_pair(idx, i));
            i++;
       }

       _atom_pos = _atomOrder.begin();
}

void Residue::find_atom(const std::string &name, std::vector<Atom*>& atom_list)
{
       atom_list.clear();
       std::string cs = CompositeIndex::getIndex(_ResName, name);
       std::pair<std::multimap<std::string, int>::iterator, std::multimap<std::string, int>::iterator> range = _atomNames.equal_range(cs);
       if (range.first == range.second) return;
       for (std::multimap<std::string, int>::iterator mpos = range.first; mpos != range.second; ++mpos) {
            atom_list.push_back(_atoms[mpos->second]);
       }
}

void Residue::find_atom_by_type(const std::string &type, std::vector<Atom*>& atom_list)
{
       atom_list.clear();
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->atom_type() != type) continue;
            atom_list.push_back((*pos));
       }
}

Atom* Residue::find_atom(const std::string &name)
{
       std::string cs = CompositeIndex::getIndex(_ResName, name);
       std::multimap<std::string, int>::iterator pos = _atomNames.find(cs);
       if (pos != _atomNames.end()) return _atoms[pos->second];
       else return NULL;
}

Atom* Residue::find_atom(const std::string &name, const std::string &alt_loc, const bool& origin_flag)
{
       if (!_atomNameAlts_Orininal.empty() && origin_flag) {
             std::string resName = _ResName;
             if (!_OrigResName.empty()) resName = _OrigResName;
             std::string cs = CompositeIndex::getIndex(resName, name, alt_loc);
             std::multimap<std::string, int>::iterator pos = _atomNameAlts_Orininal.find(cs);
             if (pos != _atomNameAlts_Orininal.end()) return _atoms[pos->second];
       }

       std::string cs = CompositeIndex::getIndex(_ResName, name, "");
       std::multimap<std::string, int>::iterator pos = _atomNameAlts.find(cs);
       if (pos != _atomNameAlts.end()) return _atoms[pos->second];

       if (!alt_loc.empty()) {
            cs = CompositeIndex::getIndex(_ResName, name, alt_loc);
            pos = _atomNameAlts.find(cs);
            if (pos != _atomNameAlts.end()) return _atoms[pos->second];
       }
       return NULL;
}

void Residue::get_atom_type()
{
       if (_atoms.empty()) return;

       bool need_update = false;
       for (std::vector<Atom*>::const_iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->atom_type().empty()) {
                 need_update = true;
                 break;
            }
       }
       if (!need_update) return;

       try {
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);
            for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
                 try {
                      const AtomFormat& atom = drug.find_atom((*pos)->pdb_atmnam());
                      (*pos)->set_atom_type(atom.atomtype());
                 } catch (const std::exception& exc) {}
            }
       } catch (const std::exception& exc) {}

       std::map<std::string, std::string> name_type;
       name_type.clear();
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (!(*pos)->atom_type().empty() &&
                 (*pos)->atom_type() != ".")
                 continue;
            else if ((*pos)->pdb_atmnam() == "1H" || (*pos)->pdb_atmnam() == "2H" || (*pos)->pdb_atmnam() == "3H")
                 (*pos)->set_atom_type("H");
            else if ((*pos)->pdb_atmnam() == "1D" || (*pos)->pdb_atmnam() == "2D" || (*pos)->pdb_atmnam() == "3D")
                 (*pos)->set_atom_type("D");
            else name_type.insert(std::make_pair((*pos)->pdb_atmnam(), ""));
       }

       if (name_type.empty()) return;

       Element::getAtomSymbols(name_type);
       bool is_aa = false;
       if (_ccDic->find_residue_type(_ResName) == "ATOMP") is_aa = true;

       std::vector<unsigned int> check_list;
       check_list.clear();
       for (unsigned int i = 0; i < _atoms.size(); ++i) {
            if (!_atoms[i]->atom_type().empty() && _atoms[i]->atom_type() != ".") continue;
            std::map<std::string, std::string>::iterator mpos = name_type.find(_atoms[i]->pdb_atmnam());
            if (mpos == name_type.end()) continue;
            if ((is_aa && (mpos->second == "HE" || mpos->second == "HF" || mpos->second == "HG")) ||
                 mpos->second == "HO") mpos->second = "H";
            _atoms[i]->set_atom_type(mpos->second);
            if (mpos->second == "HE" || mpos->second == "HF" || mpos->second == "HG") check_list.push_back(i);
       }

       if (check_list.empty()) return;

       for (std::vector<unsigned int>::const_iterator pos = check_list.begin(); pos != check_list.end(); ++pos) {
            double dist = 10000.0;
            for (unsigned int i = 0; i < _atoms.size(); ++i) {
                 if (_atoms[i]->atom_type() == "H" || _atoms[i]->atom_type() == "D" || _atoms[i]->atom_type() == "HE" ||
                     _atoms[i]->atom_type() == "HF" || _atoms[i]->atom_type() == "HG" || *pos == i) continue;

                 if (!_atoms[*pos]->alt_loc().empty() && !_atoms[i]->alt_loc().empty() &&
                     _atoms[*pos]->alt_loc() != _atoms[i]->alt_loc()) continue;
                 double distance = cal_distance(_atoms[*pos], _atoms[i]);
                 if (distance < dist) dist = distance;
            }
            if (dist < 1.6501) _atoms[*pos]->set_atom_type("H");
       }
}

void Residue::reorder_atoms()
{
       if (_atoms.empty()) return;
       if (_ResName == "UNL") return;

       try {
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);
            std::string residueType = _ccDic->find_residue_type(_ResName);

            int heavy_atom = drug.nheavyatoms();
            int extra_atom = drug.natoms() + 4;

            _atomOrder.clear();
            IntegerStringOrder idx;
            for (unsigned i = 0; i < _atoms.size(); ++i) {
                 std::string atom_name = _atoms[i]->pdb_atmnam();
                 int pos = drug.find_atom_pos(atom_name);
                 if (pos < 0 && _atoms[i]->is_hydrogen()) {
                      for (unsigned int k = 0; k < atom_name.size(); ++k) {
                           if (atom_name[k] == 'D') {
                                atom_name[k] = 'H';
                                break;
                           }
                      }
                      pos = drug.find_atom_pos(atom_name);
                 }
                 if (pos >= 0) {
                      if (_atoms[i]->atom_type() == "H" || _atoms[i]->atom_type() == "D") {
                           idx.set_int_value(pos+3);
                      } else idx.set_int_value(pos);
                      if (residueType == "ATOMP" && (_atoms[i]->pdb_atmnam() == "H2" || _atoms[i]->pdb_atmnam() == "D2")) {
                           idx.set_int_value(heavy_atom+2);
                      }
                 } else if (_atoms[i]->pdb_atmnam() == "H1" || _atoms[i]->pdb_atmnam() == "D1" || _atoms[i]->pdb_atmnam() == "H2" ||
                            _atoms[i]->pdb_atmnam() == "D2" || _atoms[i]->pdb_atmnam() == "H3" || _atoms[i]->pdb_atmnam() == "D3") {
                      if (_atoms[i]->pdb_atmnam() == "H1" || _atoms[i]->pdb_atmnam() == "D1")
                           idx.set_int_value(heavy_atom+1);
                      if (_atoms[i]->pdb_atmnam() == "H2" || _atoms[i]->pdb_atmnam() == "D2")
                           idx.set_int_value(heavy_atom+2);
                      if (_atoms[i]->pdb_atmnam() == "H3" || _atoms[i]->pdb_atmnam() == "D3")
                           idx.set_int_value(heavy_atom+3);
                 } else {
                      idx.set_int_value(extra_atom++);
                 }
                 idx.set_str_value(_atoms[i]->alt_loc());
                 _atomOrder.insert(std::make_pair(idx, i));
            }
       } catch (const std::exception& exc) {}
}

void Residue::find_extra_H_atom()
{
       std::string atomname = _ResName + "_H"; 
       std::multimap<std::string, int>::iterator ret = _atomNames.find(atomname);

       atomname = _ResName + "_H1";
       std::multimap<std::string, int>::iterator ret1 = _atomNames.find(atomname);

       atomname = _ResName + "_H2";
       std::multimap<std::string, int>::iterator ret2 = _atomNames.find(atomname);

       atomname = _ResName + "_H3";
       std::multimap<std::string, int>::iterator ret3 = _atomNames.find(atomname);

       if (ret != _atomNames.end()) {
            if (ret1 != _atomNames.end() || ret2 != _atomNames.end() || ret3 != _atomNames.end())
                 _extras.push_back("H is extra");
            else _extras.push_back("H");
            if (ret1 != _atomNames.end() && ret2 != _atomNames.end() && ret3 != _atomNames.end())
                 _extras.push_back("(H1, H2, H3 already exist)");
            else if (ret1 != _atomNames.end() && ret2 != _atomNames.end())
                 _extras.push_back("(H1, H2 already exist)");

            else if (ret1 != _atomNames.end() && ret3 != _atomNames.end())
                 _extras.push_back("(H1, H3 already exist)");
            else if (ret2 != _atomNames.end() && ret3 != _atomNames.end())
                 _extras.push_back("(H2, H3 already exist)");
            else if (ret1 != _atomNames.end())
                 _extras.push_back("(H1 already exist)");
            else if (ret2 != _atomNames.end())
                 _extras.push_back("(H1 already exist)");
            else if (ret3 != _atomNames.end())
                 _extras.push_back("(H3 already exist)");
       }
}

void Residue::find_missing_or_extra_atoms(const int& is_connect, const int& need_hydrogen, std::set<std::string>& LeavingAtoms, std::set<std::string>&
                                          AllowedTerminalAtoms, std::set<std::string>& NotAllowedTerminalAtoms, const bool& exclude_hydrogen)
{
       if (_ResName == "UNK" || _ResName == "UNL" || _ResName == "N") return;

       _missing.clear();
       _extras.clear();

       try {
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);

            _find_missing_or_extra_atoms(is_connect, need_hydrogen, LeavingAtoms, AllowedTerminalAtoms, NotAllowedTerminalAtoms, drug, exclude_hydrogen);

            if (_extras.empty()) return;

            std::vector<ConnectFormat> variants;
            _ccDic->find_variants(_ResName, variants);
            for (unsigned int i = 0; i < variants.size(); ++i) {
                 if (variants[i].natoms() < drug.natoms()) continue;
                 _missing.clear();
                 _extras.clear();
                 _find_missing_or_extra_atoms(is_connect, need_hydrogen, LeavingAtoms, AllowedTerminalAtoms, NotAllowedTerminalAtoms,
                                              variants[i], exclude_hydrogen);
                 if (_extras.empty()) return;
            }

            _missing.clear();
            _extras.clear();
            _find_missing_or_extra_atoms(is_connect, need_hydrogen, LeavingAtoms, AllowedTerminalAtoms, NotAllowedTerminalAtoms, drug, exclude_hydrogen);
       } catch (const std::exception& exc) {}
}

void Residue::refine_missing_atoms(const std::set<std::string>& bonded_atoms)
{
       if (_missing.empty()) return;
       try {
            std::set<std::string> connected_atom_set;
            connected_atom_set.clear();
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);
            std::map<std::string, std::vector<std::string> > LinkedAtoms = drug.getLinkedAtoms();
            for (std::set<std::string>::const_iterator spos = bonded_atoms.begin(); spos != bonded_atoms.end(); ++spos) {
                 std::map<std::string, std::vector<std::string> >::const_iterator mpos = LinkedAtoms.find(*spos);
                 if (mpos == LinkedAtoms.end()) continue;
                 for (std::vector<std::string>::const_iterator vpos = mpos->second.begin(); vpos != mpos->second.end(); ++vpos) {
                      connected_atom_set.insert(*vpos);
                 }
            }
            std::vector<std::string> true_missing;
            true_missing.clear();
            for (std::vector<std::string>::const_iterator vpos = _missing.begin(); vpos != _missing.end(); ++vpos) {
                 if (connected_atom_set.find(*vpos) != connected_atom_set.end()) continue;
                 true_missing.push_back(*vpos);
            }
            _missing = true_missing;
       } catch (const std::exception& exc) {}
}

void Residue::_find_missing_or_extra_atoms(const int& is_connect, const int& need_hydrogen, std::set<std::string>& LeavingAtoms,
                                           std::set<std::string>& AllowedTerminalAtoms, std::set<std::string>& NotAllowedTerminalAtoms,
                                           const ConnectFormat& drug, const bool& exclude_hydrogen)
{
       std::set<std::string> match_index;
       match_index.clear();

       const std::vector<AtomFormat>& atoms = drug.atoms();
       for (unsigned int i = 0; i < atoms.size(); ++i) {
            if (atoms[i].leaving_atom_flag() == "Y" && NotAllowedTerminalAtoms.find(atoms[i].atomname()) != NotAllowedTerminalAtoms.end()) {
                 bool skip = true;
                 if (!(is_connect & 2) && (((_ResName == "ASP" || _ResName == "DAS") && atoms[i].atomname() == "OD2") ||
                    ((_ResName == "GLU" || _ResName == "DGL") && atoms[i].atomname() == "OE2"))) skip = false;
                 else if (!(is_connect & 1) && atoms[i].atomname() == "OXT" && (_ResName == "ASP" || _ResName == "DAS" ||
                     _ResName == "DGL" || _ResName == "GLU")) skip = false;
                 else if (atoms[i].atomname() == "OXT" && !SeqCodeUtil::is_standard_aa_residue(_ResName)) skip = false;
                 if (skip) continue;
            }

            std::string cs = CompositeIndex::getIndex(_ResName, atoms[i].atomname());
            std::multimap<std::string, int>::iterator pos = _atomNames.find(cs);
            if (pos == _atomNames.end() && atoms[i].atomtype() == "H") {
                 std::string atomname = atoms[i].atomname();
                 for (unsigned int j = 0; j < atomname.size(); ++j) {
                      if (atomname[j] == 'H') {
                           atomname[j] = 'D';
                           break;
                      }
                 }
                 cs = CompositeIndex::getIndex(_ResName, atomname);
                 pos = _atomNames.find(cs);
            }

            if (pos != _atomNames.end()) match_index.insert(pos->first);
            else {
                 if (atoms[i].ishydrogen() && !need_hydrogen)
                      continue;
                 else if (LeavingAtoms.find(atoms[i].atomname()) != LeavingAtoms.end())
                      continue;
                 else if (is_connect && atoms[i].leaving_atom_flag() == "Y") {
                      bool skip = true;
                      if (!(is_connect & 2) && (((_ResName == "ASP" || _ResName == "DAS") && atoms[i].atomname() == "OD2") ||
                         ((_ResName == "GLU" || _ResName == "DGL") && atoms[i].atomname() == "OE2"))) skip = false;
                      if (!(is_connect & 1) && atoms[i].atomname() == "OXT" && (_ResName == "ASP" || _ResName == "DAS" ||
                          _ResName == "DGL" || _ResName == "GLU")) skip = false;
                      if (skip) continue;
                 } else if ((_ResName == "HOH" || _ResName == "DOD") &&
                             atoms[i].ishydrogen()) continue;

                 _missing.push_back(atoms[i].atomname());
            }
       }

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (exclude_hydrogen && (*pos)->is_hydrogen()) continue;
            if (match_index.find((*pos)->getAtomIndex()) != match_index.end()) continue;

            if (NotAllowedTerminalAtoms.find((*pos)->pdb_atmnam()) != NotAllowedTerminalAtoms.end()) {
                 _extras.push_back((*pos)->pdb_atmnam());
                 continue;
            }

            int p = drug.find_atom_pos((*pos)->pdb_atmnam());
            if (p >= 0) continue;

            std::string atomname = (*pos)->pdb_atmnam();
            if ((*pos)->is_hydrogen() && (*pos)->atom_type() == "D") {
                 for (unsigned int j = 0; j < atomname.size(); ++j) {
                      if (atomname[j] == 'D') {
                           atomname[j] = 'H';
                           break;
                      }
                 }
                 p = drug.find_atom_pos(atomname);
                 if (p >= 0) continue;
            }

            if (AllowedTerminalAtoms.find(atomname) !=
                AllowedTerminalAtoms.end()) continue;

            _extras.push_back((*pos)->pdb_atmnam());
       }
}

void Residue::find_atom_type_mismatch(const int& Mol_ID)
{
       if (_ResName == "UNL" || _ResName == "N") return;

       try {
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);

            char buffer[200];
            for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
                 try {
                      const AtomFormat& atom = drug.find_atom((*pos)->pdb_atmnam());
                      if ((*pos)->atom_type() == atom.atomtype()) continue;
     
                      std::string Chain_ID = (*pos)->pdb_chnid();
                      if (Chain_ID.empty()) Chain_ID = "  ";
                      if (Mol_ID < 0)
                           sprintf(buffer, FORMAT1, (*pos)->pdb_resnam().c_str(), Chain_ID.c_str(), (*pos)->pdb_resnum().c_str(),
                                   (*pos)->pdb_atmnam().c_str(), (*pos)->atom_type().c_str(), atom.atomtype().c_str());
                      else sprintf(buffer, FORMAT2, (*pos)->pdb_resnam().c_str(), Mol_ID, Chain_ID.c_str(), (*pos)->pdb_resnum().c_str(),
                                   (*pos)->pdb_atmnam().c_str(), (*pos)->atom_type().c_str(), atom.atomtype().c_str());
                      _messageIo->insertMessage("atom_type", "label", "true");
                      _messageIo->insertMessage("atom_type_print", "error", buffer);
                 } catch (const std::exception& exc) {}
            }
       } catch (const std::exception& exc) {}
}

void Residue::check_hydrogen_bond_distance(const int& Mol_ID)
{
       if (_ResName == "UNK" || _ResName == "UNL" || _ResName == "N") return;

       try {
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);

            bool is_cys = false;
            if (_ResName == "CYS") is_cys = true;

            std::vector<Atom*> atom_list;
            for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
                 if (!(*pos)->is_hydrogen()) continue;

                 bool is_hg = false;
                 if (is_cys && ((*pos)->pdb_atmnam() == "HG" || ((*pos)->atom_type() == "D" || (*pos)->pdb_atmnam() == "DG"))) is_hg = true;

                 int ipos = drug.find_atom_pos((*pos)->pdb_atmnam());
                 if (ipos < 0 && (*pos)->atom_type() == "D") {
                      std::string atomname = (*pos)->pdb_atmnam();
                      for (unsigned int j = 0; j < atomname.size(); ++j) {
                           if (atomname[j] == 'D') {
                                atomname[j] = 'H';
                                break;
                           }
                      }
                      ipos = drug.find_atom_pos(atomname);
                 }
                 if (ipos < 0) continue;

                 const std::vector<std::vector<std::string> >& bonds = drug.find_atom(ipos).bonds();
                 if (bonds.empty()) continue;

                 find_atom(bonds[0][0], atom_list);
                 for (unsigned int k = 0; k < atom_list.size(); ++k) {
                      if (!(*pos)->alt_loc().empty() && !atom_list[k]->alt_loc().empty() && (*pos)->alt_loc() != atom_list[k]->alt_loc()) continue;

                      double distance = cal_distance((*pos), atom_list[k]);
                      if (distance >= 0.6 && distance <= 1.2) continue;
                      if (is_hg && distance >= 0.6 && distance <= 1.4) continue;

                      _wrong_connectivity = true;
                      std::string cs = "The bond distance is " + FloatToString(distance, 0, 3) + " between '" + (*pos)->atmtype() + " "
                                     + (*pos)->pdb_resnam() + " " + (*pos)->pdb_chnid() + " " + (*pos)->pdb_resnum() + (*pos)->ins_code() + "' and '"
                                     + atom_list[k]->atmtype() + " " + atom_list[k]->pdb_resnam() + " " + atom_list[k]->pdb_chnid() + " "
                                     + atom_list[k]->pdb_resnum() + atom_list[k]->ins_code() + "'";
                      if (Mol_ID >= 0) cs += " of Model " + String::IntToString(Mol_ID);
                      cs += ".";
                      _messageIo->insertMessage("covalent_h_bond", "label", "true");
                      _messageIo->insertMessage("covalent_h_bond_print", "warning", cs);
                 }
            }
       } catch (const std::exception& exc) {}
}

void Residue::find_wrong_alt_id(const int& Mol_ID, std::string& c_9999_records)
{
       // adding check '*' in atom name
       char buffer[200];
       std::string record;

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->is_9999_flag()) {
                 _writeAtomRecord(*pos, record);
                 c_9999_records += record;
            }
            if ((*pos)->alt_loc().empty()) continue;
            if (((*pos)->alt_loc()[0] >= 'A' && (*pos)->alt_loc()[0] <= 'Z') || ((*pos)->alt_loc()[0] >= 'a' && (*pos)->alt_loc()[0] <= 'z')) continue;
            std::string Chain_ID = (*pos)->pdb_chnid();
            if (Chain_ID.empty()) Chain_ID = " ";
            if (Mol_ID < 0)
                 sprintf(buffer, FORMAT3, (*pos)->pdb_atmnam().c_str(), _ResName.c_str(), Chain_ID.c_str(), (*pos)->pdb_resnum().c_str(),
                        (*pos)->x(), (*pos)->y(), (*pos)->z(), (*pos)->alt_loc().c_str());
            else sprintf(buffer, FORMAT4, Mol_ID, (*pos)->pdb_atmnam().c_str(), _ResName.c_str(), Chain_ID.c_str(), (*pos)->pdb_resnum().c_str(),
                        (*pos)->x(), (*pos)->y(), (*pos)->z(), (*pos)->alt_loc().c_str());
            _messageIo->insertMessage("alt_code", "label", "true");
            _messageIo->insertMessage("alt_code_print", "warning", buffer);
       }
}

void Residue::check_same_coordinates(const int& Mol_ID)
{
       for (unsigned i = 0; i < _atoms.size() - 1; ++i) {
            for (unsigned j = i + 1; j < _atoms.size(); ++j) {
                 if (_atoms[i]->alt_loc().empty() || _atoms[j]->alt_loc().empty() || _atoms[i]->alt_loc() == _atoms[j]->alt_loc() ||
                    (_atoms[i]->alt_loc() != _atoms[j]->alt_loc() && _atoms[i]->pdb_atmnam() == _atoms[j]->pdb_atmnam())) {
                      float dx = fabs(_atoms[i]->x() - _atoms[j]->x());
                      float dy = fabs(_atoms[i]->y() - _atoms[j]->y());
                      float dz = fabs(_atoms[i]->z() - _atoms[j]->z());
                      float db = fabs(atof(_atoms[i]->t_fct().c_str()) - atof(_atoms[j]->t_fct().c_str()));
                      if (dx < 0.001 && dy < 0.001 && dz < 0.001 && db < 0.001)
                           _writeErrorMessage(_atoms[i], _atoms[j], "same_coor", "Atoms with same coordinates:\n", Mol_ID);
                 } 
            }
       }
}

bool Residue::is_zero_occupancy_residue(std::vector<std::vector<std::string> > &atom_list)
{
       atom_list.clear();

       std::vector<std::string> data;

       int total_count = 0;
       int zero_count = 0;
       std::set<std::string> UniIndex;
       UniIndex.clear();
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            total_count++;
            if (fabs(atof((*pos)->occ().c_str())) < ZERO_OCCUPANCY) {
                 zero_count++;
                 if (!(*pos)->is_hydrogen()) {
                      std::string cs = (*pos)->pdb_atmnam() + "_" + (*pos)->alt_loc();
                      if (UniIndex.find(cs) == UniIndex.end()) {
                           UniIndex.insert(cs);
                           data.clear();
                           data.push_back((*pos)->pdb_atmnam());
                           data.push_back((*pos)->alt_loc());
                           atom_list.push_back(data);
                      }
                 }
            }
       }

       if (zero_count && zero_count == total_count)
            return true;
       else return false;
}

bool Residue::is_zero_occupancy_ligand()
{
       int total_count = 0;
       int zero_count = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            total_count++;
            if (fabs(atof((*pos)->occ().c_str())) < ZERO_OCCUPANCY_LIGAND) zero_count++;
       }

       if (zero_count && ((zero_count * 10) >= (total_count * 8)))
            return true;
       else return false;
}

bool Residue::changed_zero_occupancy()
{
       bool changed = false;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (fabs(atof((*pos)->occ().c_str())) < ZERO_OCCUPANCY) {
                 (*pos)->setValue("1.0", 11);
                 changed = true;
            }
       }

       return changed;
}

int Residue::check_occupancy(const int& Mol_ID)
{
       int zero_occupancy_count = 0;

       if (_atoms.empty()) return zero_occupancy_count;

       float occupancy = 0;
       std::string name = "XXXX";
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            float occ = atof((*pos)->occ().c_str());
            if (name != (*pos)->pdb_atmnam()) {
                 if (occupancy > 1.0) _writeOccupancyMessage(name, occupancy, Mol_ID);
                 name = (*pos)->pdb_atmnam();
                 occupancy = 0;
            }
            occupancy += occ;
            if (fabs(occ) < ZERO_OCCUPANCY) zero_occupancy_count++;
            if (occ < 0) _writeOccupancyMessage((*pos)->pdb_atmnam(), occ, Mol_ID);

       }
       if (occupancy > 1.0) _writeOccupancyMessage(name, occupancy, Mol_ID);

       return zero_occupancy_count;
}

bool Residue::check_occupancy_and_altloc(const int& Mol_ID)
{
       bool error_free_flag = true;

       std::map<std::string, int> altloc_count_mapping;
       altloc_count_mapping.clear();

       float occupancy = 0;
       std::string name = "____XXXX";
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            std::string idx = (*pos)->pdb_atmnam() + "_" + (*pos)->alt_loc();
            std::map<std::string, int>::iterator
                mpos = altloc_count_mapping.find(idx);
            if (mpos != altloc_count_mapping.end())
                 mpos->second++;
            else altloc_count_mapping.insert(std::make_pair(idx, 1));

            float occ = atof((*pos)->occ().c_str());
            if (name != (*pos)->pdb_atmnam()) {
                 if (occupancy > 1.0) {
                      _writeOccupancyError(name, occupancy, Mol_ID);
                      error_free_flag = false;
                 }
                 name = (*pos)->pdb_atmnam();
                 occupancy = 0;
            }
            occupancy += occ;

       }
       if (occupancy > 1.0) {
            _writeOccupancyError(name, occupancy, Mol_ID);
            error_free_flag = false;
       }

       std::vector<std::string> data;
       std::string error;
       for (std::map<std::string, int>::const_iterator mpos = altloc_count_mapping.begin(); mpos != altloc_count_mapping.end(); ++mpos) {
            if (mpos->second < 2) continue; 
            get_wordarray_delimit_by_string(data, mpos->first, "_");

            error.clear();
            if (Mol_ID >= 0) error += "In model " + String::IntToString(Mol_ID) + ",";

            error += "Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + data[0];
            if (!data[1].empty()) error += "[InsCode:" + data[1] + "]";
            error += "' repeats " + String::IntToString(mpos->second) + " times.\n";
            _logIo->message(error.c_str());
            error_free_flag = false;
       }

       return error_free_flag;
}

std::string Residue::check_b_factor(int& atom_count, int& high_b_factor_count, int& low_b_factor_count, int& negative_b_factor_count)
{
       if (_atoms.empty()) return "";

       bool has_same_b_factor = true;
       float t_fct = -1000.0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            atom_count++;

            float fct = atof((*pos)->t_fct().c_str());
            if (fct > HIGH_B_FACTOR) high_b_factor_count++;
            if (fct < LOW_B_FACTOR) low_b_factor_count++;
            if (fct < ZERO_B_FACTOR) negative_b_factor_count++;
            if (t_fct < -999.0) t_fct = fct;
            if (fabs(fct - t_fct) > 0.01) has_same_b_factor = false;

       }

       if (has_same_b_factor) return FloatToString(t_fct, 0, 2);

       return "";
}

void Residue::check_occupancy_and_b_factor(std::vector<std::string>& occ_list, std::vector<std::string>& fct_list)
{
       if (_atoms.empty()) return;

       float occupancy = 0;
       std::string name = "XXXX";
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            float occ = atof((*pos)->occ().c_str());
            if (occ < 0) {
                 occ_list.push_back("Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + (*pos)->pdb_atmnam()
                                  + " " + (*pos)->alt_loc() + "' has occupancy " + FloatToString(occ, 0, 2) + ".");
            }
            if (name != (*pos)->pdb_atmnam()) {
                 if (occupancy > 1.0) {
                      occ_list.push_back("Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + name
                                       + "' has occupancy " + FloatToString(occupancy, 0, 2) + ".");
                 }
                 name = (*pos)->pdb_atmnam();
                 occupancy = 0;
            }
            occupancy += occ;

            float fct = atof((*pos)->t_fct().c_str());
            if (fct < ZERO_B_FACTOR) {
                 fct_list.push_back("Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + (*pos)->pdb_atmnam()
                                  + " " + (*pos)->alt_loc() + "' has temperature factor " + FloatToString(fct, 0, 2) + ".");
            }
       }
       if (occupancy > 1.0) {
            occ_list.push_back("Atom '" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + " " + name
                             + "' has occupancy " + FloatToString(occupancy, 0, 2) + ".");
       }
}

void Residue::CorresInfoChecking(bool& high_b_factor, std::list<Atom*>& zero_atom_list, std::list<Atom*>& large_atom_list)
{
       high_b_factor = false; 
       if (_atoms.empty()) return;

       int atom_count = 0;
       int high_b_factor_count = 0;
       float occupancy = 0;
       std::string name = "XXXX";

       std::list<Atom*> atom_list;
       atom_list.clear();

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            atom_count++;
            if (atof((*pos)->t_fct().c_str()) > HIGH_B_FACTOR) high_b_factor_count++;

            float occ = atof((*pos)->occ().c_str());
            if (fabs(occ) < ZERO_OCCUPANCY) zero_atom_list.push_back(*pos);

            if (name != (*pos)->pdb_atmnam()) {
                 if (occupancy > 1.0 && !atom_list.empty()) {
                      for (std::list<Atom*>::const_iterator lpos = atom_list.begin(); lpos != atom_list.end(); ++lpos) {
                           large_atom_list.push_back(*lpos);
                      }
                 }
                 name = (*pos)->pdb_atmnam();
                 occupancy = 0;
                 atom_list.clear();
            }
            occupancy += occ;
            atom_list.push_back(*pos);
       }
       if (occupancy > 1.0 && !atom_list.empty()) {
            for (std::list<Atom*>::const_iterator lpos = atom_list.begin(); lpos != atom_list.end(); ++lpos) {
                 large_atom_list.push_back(*lpos);
            }
       }

       if (atom_count == high_b_factor_count) high_b_factor = true;
}

bool Residue::change_water_nomenclature()
{
       bool changed = false;
       bool changed_residue = false;
       if (_ResName == "H2O" || _ResName == "WAT" || _ResName == "TIP" || _ResName == "WTR" || _ResName == "OH2") {
            _ResName = "HOH";
            changed_residue = true;
            changed = true;
       }
       if (_ResName != "HOH" && _ResName != "DOD") return  changed_residue;

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (changed) {
                 (*pos)->set_restype(_ResName);
                 (*pos)->set_pdb_resnam(_ResName);
            }

            std::string p = (*pos)->atmtype();
            if (p == "OW" || p == "WAT" || p == "OH2" || p == "WTR" || p == "H2O" || p == "W" || p[0] == 'O') p = "O";
            else if (p == "1H") p = "H1";
            else if (p == "2H") p = "H2";
            else if (p == "1D") p = "D1";
            else if (p == "2D") p = "D2";
            if (p != (*pos)->atmtype()) {
                 (*pos)->set_atmtype(p);
                 (*pos)->set_pdb_atmnam(p);
                 changed = true;
            }
       }

       if (changed) {
            if (_atomNameAlts_Orininal.empty()) _atomNameAlts_Orininal = _atomNameAlts;
            _UpdateIndices();
       }
       return changed_residue;
}

void Residue::correction_special_na_name(const bool& first_flag, const int& Mol_ID)
{
       std::string resname = SeqCodeUtil::GetStandardNaCode(_ResName);
       if (resname.empty()) {
            if (find_atom("O2'") || find_atom("O2*")) {
                 if (_ResName == "ADE")
                      resname = "A";
                 else if (_ResName == "CYT")
                      resname = "C";
                 else if (_ResName == "GUA")
                      resname = "G";
            } else {
                 if (_ResName == "ADE")
                      resname = "DA";
                 else if (_ResName == "CYT")
                      resname = "DC";
                 else if (_ResName == "GUA")
                      resname = "DG";
            }
       }
       if (resname.empty()) return;

       try {
            const ConnectFormat& drug = _ccDic->find_drug(resname);

            std::vector<std::pair<std::string, std::string> > atomlist, bondlist;
            _createScratchGraph(Mol_ID, atomlist, bondlist);
            if (atomlist.empty() || bondlist.empty()) return;

            bool is_substructure_match = false;
            std::map<std::string, std::string> mapping_list;
            std::vector<std::pair<std::string, std::string> > ref_a_list = drug.getAtomList();
            std::vector<std::pair<std::string, std::string> > ref_b_list = drug.getBondList();
            GraphMatch::GetMatch(ref_a_list, ref_b_list, atomlist, bondlist, mapping_list,
                                 is_substructure_match, TARGET_TO_REF, true);
            if (mapping_list.empty()) return;

            if (is_substructure_match) {
                 if (first_flag) {
                      bool has_sugar_atom = false;
                      if ((find_atom("C1'") || find_atom("C1*")) && (find_atom("C2'") || find_atom("C2*")) && (find_atom("C3'") || find_atom("C3*")) &&
                          (find_atom("C4'") || find_atom("C4*")) && (find_atom("C5'") || find_atom("C5*")) && (find_atom("O3'") || find_atom("O3*")) &&
                          (find_atom("O4'") || find_atom("O4*")) && (find_atom("O5'") || find_atom("O5*"))) has_sugar_atom = true;
                      if (!has_sugar_atom) return;
                 } else if (!find_atom("P")) return;
            }

            _ResName = resname;
            _chain_type = "ATOMN";
            for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
                 std::map<std::string, std::string>::const_iterator mpos = mapping_list.find((*pos)->pdb_atmnam());
                 if (mpos != mapping_list.end()) {
                      (*pos)->set_atmtype(mpos->second);
                      (*pos)->set_pdb_atmnam(mpos->second);
                 }
                 (*pos)->set_restype(_ResName);
                 (*pos)->set_pdb_resnam(_ResName);
            }

            if (_atomNameAlts_Orininal.empty()) _atomNameAlts_Orininal = _atomNameAlts;
            _UpdateIndices();
       } catch (const std::exception& exc) {}
}

void Residue::AtomNameUpperCase()
{
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            (*pos)->UpperCase();
       }
       _atomNameAlts_Orininal.clear();
       _UpdateIndices();
       reorder_atoms();
}

void Residue::correction_atom_name(const int& Mol_ID)
{
       if (_ResName == "UNL") return;

       if (_uniqueAtomNames.size() == 1) {
            try {
                 const ConnectFormat& drug = _ccDic->find_drug(_ResName);
                 const std::vector<AtomFormat>& atoms = drug.atoms();
                 if (atoms.size() == 1) {
                      for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
                           (*pos)->set_atmtype(atoms[0].atomname());
                           (*pos)->set_pdb_atmnam(atoms[0].atomname());
                           (*pos)->set_atom_type(atoms[0].atomtype());
                      }
                      if (_atomNameAlts_Orininal.empty()) _atomNameAlts_Orininal = _atomNameAlts;
                      _UpdateIndices();
                 }
            } catch (const std::exception& exc) {}
            return;
       }

       std::vector<ConnectFormat> drug_list;
       _ccDic->find_drugs(_ResName, drug_list);
       if (drug_list.empty()) return;

       std::vector<std::pair<std::string, std::string> > atomlist, bondlist;
       _createScratchGraph(Mol_ID, atomlist, bondlist);
       if (atomlist.empty() || bondlist.empty()) return;

       std::map<std::string, std::string> atomtype;
       atomtype.clear();
       for (std::vector<std::pair<std::string, std::string> >::iterator pos = atomlist.begin(); pos != atomlist.end(); ++pos) {
            atomtype.insert(std::make_pair(pos->second, pos->first));
       }

       std::multimap<unsigned int, std::pair<int, std::map<std::string, std::string> > > all_mapping_list;
       all_mapping_list.clear();
       unsigned int max_mapped_vertex = 0;
       // record all mapping from component dictionary (including variants)
       // sort by matched vertex numbers

       bool is_substructure_match = false;
       std::map<std::string, std::string> list;
       for (std::vector<ConnectFormat>::const_iterator dpos = drug_list.begin(); dpos != drug_list.end(); ++dpos) {
            std::vector<std::pair<std::string, std::string> > ref_a_list = dpos->getAtomList();
            std::vector<std::pair<std::string, std::string> > ref_b_list = dpos->getBondList();
            GraphMatch::GetMatch(ref_a_list, ref_b_list, atomlist, bondlist, list, is_substructure_match);
            if (list.empty()) continue;

            bool bad_mapping = false;
            int count = 0;
            for (std::map<std::string, std::string>::const_iterator mpos = list.begin(); mpos != list.end(); ++mpos) {
                 std::string aName = mpos->first;
                 std::string eType = ""; 
                 std::map<std::string, std::string>::const_iterator mpos1 = atomtype.find(aName);
                 if (mpos1 != atomtype.end()) eType = mpos1->second;
                 if (eType == "D") {
                      for (unsigned int k = 0; k < aName.size(); k++) {
                           if (aName[k] == 'D') {
                                aName[k] = 'H';
                                break;
                           }
                      }

                 }

                 std::string cs = CompositeIndex::getIndex(_ResName, mpos->second);
                 if (list.find(mpos->second) == list.end() && eType != "D" &&  _atomNames.find(cs) != _atomNames.end()) {
                      bad_mapping = true;
                 }
                 if (aName != mpos->second) count++; 
            }
            if (bad_mapping) continue;

            if (list.size() == atomlist.size() && count == 0) return;

            if (list.size() > max_mapped_vertex) max_mapped_vertex = list.size();
            all_mapping_list.insert(std::make_pair(list.size(), std::make_pair(count, list)));
       }

       std::multimap<unsigned int, std::map<std::string, std::string> > selected_mapping_list;
       selected_mapping_list.clear();
       // select mapping with maximum matched vertex numbers
       // sort by the number of atom name difference
       if (!all_mapping_list.empty()) {
            std::pair<std::multimap<unsigned int, std::pair<int, std::map<std::string, std::string> > >::iterator,
                      std::multimap<unsigned int, std::pair<int, std::map<std::string, std::string> > >::iterator>
                  range = all_mapping_list.equal_range(max_mapped_vertex);
            for (std::multimap<unsigned int, std::pair<int, std::map<std::string, std::string> > >::iterator
                 mmpos = range.first; mmpos != range.second; ++mmpos) {
                 selected_mapping_list.insert(std::make_pair(mmpos->second.first, mmpos->second.second));
            }
       }
       if (selected_mapping_list.empty()) {
            std::string cs = "In model " + String::IntToString(Mol_ID) + ", Can't find match for residue (";
            if (Mol_ID < 0) cs = "Can't find match for residue (";
            cs += GetFirstAtom()->pdb_chnid() + " " + GetFirstAtom()->pdb_resnam() + " " + GetFirstAtom()->pdb_resnum() + ").";
            _messageIo->insertMessage("residue_match_print", "warning", cs);
            cs += "\n";
            _logIo->message(cs.c_str());
            return;
       }

       std::multimap<unsigned int, std::map<std::string, std::string> >::iterator mmpos = selected_mapping_list.begin();
       if (mmpos->first == 0) return;

       for (std::map<std::string, std::string>::iterator mpos = mmpos->second.begin(); mpos != mmpos->second.end(); ++mpos) {
            std::string eType = "";
            std::map<std::string, std::string>::const_iterator mpos1 = atomtype.find(mpos->first);
            if (mpos1 != atomtype.end()) eType = mpos1->second;
            _update_atom_nomenclature(mpos->first, mpos->second, eType);

            // handle D/H mixed alternative conformation
            if (eType == "D") {
                 std::string aName = mpos->first;
                 for (unsigned int k = 0; k < aName.size(); k++) {
                      if (aName[k] == 'D') {
                           aName[k] = 'H';
                           break;
                      }
                 }
                 _update_atom_nomenclature(aName, mpos->second, "H");
            } else if (eType == "H") {
                 std::string aName = mpos->first;
                 for (unsigned int k = 0; k < aName.size(); k++) {
                      if (aName[k] == 'H') {
                           aName[k] = 'D';
                           break;
                      }
                 }
                 _update_atom_nomenclature(aName, mpos->second, "D");
            }
       }

       if (_atomNameAlts_Orininal.empty()) _atomNameAlts_Orininal = _atomNameAlts;
       _UpdateIndices();
       reorder_atoms();
}

void Residue::_update_atom_nomenclature(const std::string& old_name, const std::string& new_name, const std::string& atom_type)
{
       std::string cs = CompositeIndex::getIndex(_ResName, old_name);
       std::pair<std::multimap<std::string, int>::iterator, std::multimap<std::string, int>::iterator> range = _atomNames.equal_range(cs);
       if (range.first == range.second) return;

       std::string name = new_name;
       if (atom_type == "D") {
            for (unsigned int k = 0; k < name.size(); k++) {
                 if (name[k] == 'H') {
                      name[k] = 'D';
                      break;
                 }
            }
       }

       for (std::multimap<std::string, int>::iterator mpos = range.first; mpos != range.second; ++mpos) {
            _atoms[mpos->second]->set_atmtype(name);
            _atoms[mpos->second]->set_pdb_atmnam(name);
       }
}

void Residue::update_nomenclature(const std::string &type, const std::string& res_num)
{
       _res_no = res_num;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (!type.empty()) (*pos)->set_type(type);
            (*pos)->set_chnid(_chnid);
            (*pos)->set_resnum(res_num);
       }
}

void Residue::update_nomenclature(const std::string &type, const std::string &chain_id, const std::string& res_num)
{
       _chnid = chain_id;
       _res_no = res_num;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (!type.empty()) (*pos)->set_type(type);
            (*pos)->set_chnid(chain_id);
            (*pos)->set_resnum(res_num);
       }
}

void Residue::update_nomenclature(const std::string &type, const std::string &chain_id, const std::string& res_num, const std::string &pdb_chainid,
                                  const std::string& pdb_res_num, const std::string& ins_code)
{
       _chnid = chain_id;
       _res_no = res_num;
       _pdb_chnid = pdb_chainid;
       _pdb_res_no = pdb_res_num;
       _ins_code = ins_code;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (!type.empty()) (*pos)->set_type(type);
            (*pos)->set_chnid(chain_id);
            (*pos)->set_resnum(res_num);
            (*pos)->set_pdb_chnid(pdb_chainid);
            (*pos)->set_pdb_resnum(pdb_res_num);
            (*pos)->set_ins_code(ins_code);
       }
}

void Residue::update_nomenclature(const std::map<std::string, std::string>& mapping)
{
       if (mapping.empty()) return;

       bool chainid_flag = false;
       bool number_flag = false;
       bool inscode_flag = false;
       std::map<std::string, std::string>::const_iterator mpos = mapping.find("chainid");
       if (mpos != mapping.end()) {
            _pdb_chnid = mpos->second;
            chainid_flag = true;
       }

       mpos = mapping.find("number");
       if (mpos != mapping.end()) {
            _pdb_res_no = mpos->second;
            number_flag = true;
       }

       mpos = mapping.find("inscode");
       if (mpos != mapping.end()) {
            _ins_code = mpos->second;
            inscode_flag = true;
       }

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (chainid_flag) (*pos)->set_pdb_chnid(_pdb_chnid);
            if (number_flag) (*pos)->set_pdb_resnum(_pdb_res_no);
            if (inscode_flag) (*pos)->set_ins_code(_ins_code);
       }
}

void Residue::correction_name(const std::string &resname)
{
       if (_ResName == resname) return;

       _ResName = resname;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            (*pos)->set_restype(_ResName);
            (*pos)->set_pdb_resnam(_ResName);
       }
       _atomNameAlts_Orininal.clear();
       _UpdateIndices();
       reorder_atoms();

}

void Residue::convert_between_orthogonal_and_fractional(CrySymmetry& crySymm, const int& iflag)
{
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            COORD coord = (*pos)->orig();
            crySymm.ndb_trans_coord_3by3(coord, iflag);
            (*pos)->set_orig(coord);
       }
}

void Residue::symmetry_operations(CrySymmetry& crySymm, const int& lx, const int& ly, const int& lz, const NDBSYMMETRY& op, const bool& not_moving)
{
       bool exist_anisou = false;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->has_anisou()) {
                 exist_anisou = true;
                 break;
            }
       }
       if (exist_anisou && not_moving) return;

       COORD coord;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            crySymm.do_symmetry_operation(coord, (*pos)->orig(), op);
            coord.x += (float) lx;
            coord.y += (float) ly;
            coord.z += (float) lz;
            (*pos)->set_orig(coord);
       }
}

void Residue::transformation(double m[4][4])
{
       COORD coord;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            matrix_rotate_translate(coord, (*pos)->orig(), m);
            (*pos)->set_orig(coord);
       }
}

void Residue::rename(const std::string& NewName)
{
       std::set<std::string> remove_atom_set;
       remove_atom_set.clear();

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            (*pos)->set_restype(NewName);
            (*pos)->set_pdb_resnam(NewName);
            if (_ResName == "ALA" && NewName == "GLY") {
                if (((*pos)->pdb_atmnam() == "CB") ||  ((*pos)->pdb_atmnam() == "HB1") || ((*pos)->pdb_atmnam() == "HB2") || ((*pos)->pdb_atmnam() == "HB3") ||
                    ((*pos)->pdb_atmnam() == "1HB") || ((*pos)->pdb_atmnam() == "2HB") || ((*pos)->pdb_atmnam() == "3HB"))
                     remove_atom_set.insert((*pos)->pdb_atmnam());
            } else if (_ResName == "ASP" && NewName == "ASN" && (*pos)->pdb_atmnam() == "OD2") {
                 (*pos)->set_atmtype("ND2");
                 (*pos)->set_pdb_atmnam("ND2");
                 (*pos)->set_atom_type("N");
            } else if (_ResName == "ASN" && NewName == "ASP" && (*pos)->pdb_atmnam() == "ND2") {
                 (*pos)->set_atmtype("OD2");
                 (*pos)->set_pdb_atmnam("OD2");
                 (*pos)->set_atom_type("O");
            } else if (_ResName == "GLU" && NewName == "GLN" && (*pos)->pdb_atmnam() == "OE2") {
                 (*pos)->set_atmtype("NE2");
                 (*pos)->set_pdb_atmnam("NE2");
                 (*pos)->set_atom_type("N");
            } else if (_ResName == "GLN" && NewName == "GLU" && (*pos)->pdb_atmnam() == "NE2") {
                 (*pos)->set_atmtype("OE2");
                 (*pos)->set_pdb_atmnam("OE2");
                 (*pos)->set_atom_type("O");
            }
       }

       std::set<long> atom_set;
       atom_set.clear();
       if (!remove_atom_set.empty()) {
            for (std::set<std::string>::const_iterator spos = remove_atom_set.begin(); spos != remove_atom_set.end(); ++spos) {
                 delete_a_atom(*spos, atom_set);
            }
       }

       _ResName = NewName;
       _atomNameAlts_Orininal.clear();
       _UpdateIndices();
       reorder_atoms();
}

void Residue::fix_n_terminal_hydrogen(const bool& first_residue)
{
       int changed = 0;

       std::string pro_H1 = "H2";
       std::string pro_H2 = "H3";
       std::string atom_H1 = "H1";
       std::string atom_H2 = "H2";
       std::string atom_H3 = "H3";

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (_ResName == "PRO") {
                 if ((*pos)->pub_atmnam() == "HT1" || (*pos)->pub_atmnam() == "1HT" || (*pos)->pub_atmnam() == "HN2" || (*pos)->pub_atmnam() == "1H" ||
                     (*pos)->pub_atmnam() == "H2") { 
                      (*pos)->set_atmtype(pro_H1);
                      (*pos)->set_pdb_atmnam(pro_H1);
                      changed++;
                 } else if ((*pos)->pub_atmnam() == "HT2" || (*pos)->pub_atmnam() == "2HT" || (*pos)->pub_atmnam() == "HN1" || (*pos)->pub_atmnam() == "2H" ||
                            (*pos)->pub_atmnam() == "H3") { 
                      (*pos)->set_atmtype(pro_H2);
                      (*pos)->set_pdb_atmnam(pro_H2);
                      changed++;
                 }
                 continue;
            }

            if ((*pos)->pub_atmnam() == "HT1" || (*pos)->pub_atmnam() == "1HT" || (*pos)->pub_atmnam() == "1HN" ||
                (*pos)->pub_atmnam() == "HN1" || (*pos)->pub_atmnam() == "1H" || (*pos)->pub_atmnam() == "H1") { 
                 (*pos)->set_atmtype(atom_H1);
                 (*pos)->set_pdb_atmnam(atom_H1);
                 changed++;
            } else if ((*pos)->pub_atmnam() == "HT2" || (*pos)->pub_atmnam() == "2HT" || (*pos)->pub_atmnam() == "2HN" ||
                       (*pos)->pub_atmnam() == "HN2" || (*pos)->pub_atmnam() == "2H" || (*pos)->pub_atmnam() == "H2") { 
                 (*pos)->set_atmtype(atom_H2);
                 (*pos)->set_pdb_atmnam(atom_H2);
                 changed++;
            } else if ((*pos)->pub_atmnam() == "HT3" || (*pos)->pub_atmnam() == "3HT" || (*pos)->pub_atmnam() == "3HN" ||
                       (*pos)->pub_atmnam() == "HN3" || (*pos)->pub_atmnam() == "3H" || (*pos)->pub_atmnam() == "H3") { 
                 (*pos)->set_atmtype(atom_H3);
                 (*pos)->set_pdb_atmnam(atom_H3);
                 changed++;
            } else if ((*pos)->pub_atmnam() == "HN" || (*pos)->pub_atmnam() == "H") {
                 (*pos)->set_atmtype(atom_H1);
                 (*pos)->set_pdb_atmnam(atom_H1);
                 changed++;
            } else if ((*pos)->pub_atmnam() == "1D" || (*pos)->pub_atmnam() == "D1" || (*pos)->pub_atmnam() == "DT1") {
                 (*pos)->set_atmtype("D1");
                 (*pos)->set_pdb_atmnam("D1");
                 changed++;
            } else if ((*pos)->pub_atmnam() == "2D" || (*pos)->pub_atmnam() == "D2" || (*pos)->pub_atmnam() == "DT2") {
                 (*pos)->set_atmtype("D2");
                 (*pos)->set_pdb_atmnam("D2");
                 changed++;
            } else if ((*pos)->pub_atmnam() == "3D" || (*pos)->pub_atmnam() == "D3" || (*pos)->pub_atmnam() == "DT3") {
                 (*pos)->set_atmtype("D3");
                 (*pos)->set_pdb_atmnam("D3");
                 changed++;
            }
       }

       if (changed == 1) {
            for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
                 if (_ResName == "PRO") {
                      if ((*pos)->pub_atmnam() == "HT1" || (*pos)->pub_atmnam() == "HN2" || (*pos)->pub_atmnam() == "1H" || (*pos)->pub_atmnam() == "H2" ||
                          (*pos)->pub_atmnam() == "HT2" || (*pos)->pub_atmnam() == "HN1" || (*pos)->pub_atmnam() == "2H" || (*pos)->pub_atmnam() == "H3") { 
                           if (first_residue) {
                                (*pos)->set_atmtype(pro_H1);
                                (*pos)->set_pdb_atmnam(pro_H1);
                           } else {
                                (*pos)->set_atmtype("H");
                                (*pos)->set_pdb_atmnam("H");
                           }
                      }
                      continue;
                 }
     
                 if ((*pos)->pub_atmnam() == "HT1" || (*pos)->pub_atmnam() == "1HT" || (*pos)->pub_atmnam() == "1HN" || (*pos)->pub_atmnam() == "HN1" ||
                     (*pos)->pub_atmnam() == "1H" || (*pos)->pub_atmnam() == "H1" || (*pos)->pub_atmnam() == "HT2" || (*pos)->pub_atmnam() == "2HT" ||
                     (*pos)->pub_atmnam() == "2HN" || (*pos)->pub_atmnam() == "HN2" || (*pos)->pub_atmnam() == "2H" || (*pos)->pub_atmnam() == "H2" ||
                     (*pos)->pub_atmnam() == "HT3" || (*pos)->pub_atmnam() == "3HT" || (*pos)->pub_atmnam() == "3HN" || (*pos)->pub_atmnam() == "HN3" ||
                     (*pos)->pub_atmnam() == "3H" || (*pos)->pub_atmnam() == "H3" || (*pos)->pub_atmnam() == "HN" || (*pos)->pub_atmnam() == "H") {
                      if (first_residue) {
                           (*pos)->set_atmtype(atom_H1);
                           (*pos)->set_pdb_atmnam(atom_H1);
                      } else {
                           (*pos)->set_atmtype("H");
                           (*pos)->set_pdb_atmnam("H");
                      }
                 }
            }
       }

       if (changed) {
            _atomNameAlts_Orininal.clear();
            _UpdateIndices();
            reorder_atoms();
       }
}

void Residue::fix_n_terminal_hydrogen_based_on_mapping()
{
       int changed = 0;

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (_ResName == "PRO") {
                 if ((*pos)->atmtype() == "HT1" || (*pos)->atmtype() == "1H") {
                      (*pos)->set_atmtype("H2");
                      (*pos)->set_pdb_atmnam("H2");
                      changed++;
                 } else if ((*pos)->atmtype() == "HT2" || (*pos)->atmtype() == "2H") { 
                      (*pos)->set_atmtype("H3");
                      (*pos)->set_pdb_atmnam("H3");
                      changed++;
                 }
                 continue;
            }

            if ((*pos)->atmtype() == "1H") {
                 (*pos)->set_atmtype("H1");
                 (*pos)->set_pdb_atmnam("H1");
                 changed++;
            } else if ((*pos)->atmtype() == "2H") {
                 (*pos)->set_atmtype("H2");
                 (*pos)->set_pdb_atmnam("H2");
                 changed++;
            } else if ((*pos)->atmtype() == "3H") {
                 (*pos)->set_atmtype("H3");
                 (*pos)->set_pdb_atmnam("H3");
                 changed++;
            } else if ((*pos)->atmtype() == "1D") {
                 (*pos)->set_atmtype("D1");
                 (*pos)->set_pdb_atmnam("D1");
                 changed++;
            } else if ((*pos)->atmtype() == "2D") {
                 (*pos)->set_atmtype("D2");
                 (*pos)->set_pdb_atmnam("D2");
                 changed++;
            } else if ((*pos)->atmtype() == "3D") {
                 (*pos)->set_atmtype("D3");
                 (*pos)->set_pdb_atmnam("D3");
                 changed++;
            }
       }

       if (changed) {
            _atomNameAlts_Orininal.clear();
            _UpdateIndices();
            reorder_atoms();
       }
}

void Residue::fix_5_terminal_hydrogen()
{
       std::string H_atom = "HO5'";

       int changed = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->pub_atmnam() == "H5T" || (*pos)->pub_atmnam() == "HO5*" || (*pos)->pub_atmnam() == "H5*" || (*pos)->pub_atmnam() == "HO5'") {
                 (*pos)->set_atmtype(H_atom);
                 (*pos)->set_pdb_atmnam(H_atom);
                 changed++;
            }
       }
       if (changed) {
            _atomNameAlts_Orininal.clear();
            _UpdateIndices();
            reorder_atoms();
       }
}

void Residue::fix_5_terminal_hydrogen_based_on_mapping()
{
       int changed = 0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if ((*pos)->atmtype() == "H5T") {
                 (*pos)->set_atmtype("HO5'");
                 (*pos)->set_pdb_atmnam("HO5'");
                 changed++;
            }
       }
       if (changed) {
            _atomNameAlts_Orininal.clear();
            _UpdateIndices();
            reorder_atoms();
       }
}

void Residue::check_connectivity(const int& Mol_ID)
{
       if (_uniqueAtomNames.size() < 2) return;

       try {
            const ConnectFormat& drug = _ccDic->find_drug(_ResName);
            const std::vector<std::vector<std::string> >& bonds = drug.bonds();
            if (bonds.size() < 2) return;

            std::string ResString = _ResName + " " + _pdb_chnid + " " + _pdb_res_no + _ins_code;
            if (Mol_ID > 0) ResString += " ( Model " + String::IntToString(Mol_ID) + ")";
            bool break_flag = false;
            std::vector<Atom*> atom1, atom2;
            for (unsigned int i = 0; i < bonds.size(); i++) {
                 find_atom(bonds[i][0], atom1);
                 find_atom(bonds[i][1], atom2);
                 for (std::vector<Atom*>::iterator pos1 = atom1.begin(); pos1 != atom1.end(); ++pos1) {
                      for (std::vector<Atom*>::iterator pos2 = atom2.begin(); pos2 != atom2.end(); ++pos2) {
                           if (!(*pos1)->alt_loc().empty() && !(*pos2)->alt_loc().empty() && (*pos1)->alt_loc() != (*pos2)->alt_loc()) continue;
                           double dist = cal_distance(*pos1, *pos2);
                           if (dist <= 3.2) continue;
     
                           // std::string cs = _ResName + " " + (*pos1)->pdb_chnid() + " " + (*pos1)->pdb_resnum();
                           // if (!(*pos1)->ins_code().empty()) cs += (*pos1)->ins_code();
                           // if (Mol_ID > 0) cs += " (" + String::IntToString(Mol_ID) + ")";
                           _messageIo->insertMessage("dissociated_residue", "warning", ResString);
                           _messageIo->insertMessage("dissociated_residue_print", "warning", "Dissociated residue: " + ResString);
                           break_flag = true;
                           break;
                      }
                      if (break_flag) break;
                 }
                 if (break_flag) break;
            }

            std::set<std::string> bond_set;
            bond_set.clear();
            for (unsigned int i = 0; i < bonds.size(); i++) {
                 bond_set.insert(bonds[i][0] + "_" + bonds[i][1]);
                 bond_set.insert(bonds[i][1] + "_" + bonds[i][0]);
            }
            for (unsigned int i = 0; i < _atoms.size() - 1; i++) {
                 for (unsigned int j = i + 1; j < _atoms.size(); j++) {
                      if (!_atoms[i]->alt_loc().empty() && !_atoms[j]->alt_loc().empty() && _atoms[i]->alt_loc() != _atoms[j]->alt_loc()) continue;
                      if ((_atoms[i]->atom_type() == "H" || _atoms[i]->atom_type() == "D") &&
                          (_atoms[j]->atom_type() == "H" || _atoms[j]->atom_type() == "D")) continue;
                      if (!BondUtil::is_a_bond(_atoms[i], _atoms[j])) continue;
                      if (bond_set.find(_atoms[i]->pdb_atmnam() + "_" + _atoms[j]->pdb_atmnam()) != bond_set.end()) continue;

                      std::string atom_i = _atoms[i]->pdb_atmnam();
                      if (!_atoms[i]->alt_loc().empty()) atom_i += "(" + _atoms[i]->alt_loc() + ")";
                      std::string atom_j = _atoms[j]->pdb_atmnam();
                      if (!_atoms[j]->alt_loc().empty()) atom_j += "(" + _atoms[j]->alt_loc() + ")";
                      std::string cs = "In '" + ResString + "' residue: incorrect bonding between atom '" + atom_i + "' and atom '" + atom_j + "'.";
                      _messageIo->insertMessage("residue_match_print", "warning", cs);
                 }
            }
       } catch (const std::exception& exc) {}
}

bool Residue::update_atom_name(const std::map<std::string, std::string>& atom_mapping, const int& Mol_ID, const bool& index_flag)
{
       bool _successful = true;
       if (atom_mapping.empty()) return _successful;

       std::map<std::string, std::string> new_atom_mapping;
       new_atom_mapping.clear();

       std::set<std::string> extra_atom_set;
       extra_atom_set.clear();
       std::vector<std::string> messages;
       messages.clear();
       for (std::vector<Atom*>::const_iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            std::string atomname = (*pos)->atmtype();
            std::map<std::string, std::string>::const_iterator mpos = atom_mapping.find(atomname);
            if (mpos != atom_mapping.end()) {
                 new_atom_mapping.insert(std::make_pair((*pos)->atmtype(), mpos->second));
                 continue;
            }
            if ((*pos)->atom_type() == "D") {
                 atomname[0] = 'H';
                 mpos = atom_mapping.find(atomname);
                 if (mpos != atom_mapping.end()) {
                      std::string mapped_name = mpos->second;
                      mapped_name[0] = 'D';
                      new_atom_mapping.insert(std::make_pair((*pos)->atmtype(), mapped_name));
                      continue;
                 }
            } else if ((*pos)->atom_type() == "H") {
                 atomname[0] = 'D';
                 mpos = atom_mapping.find(atomname);
                 if (mpos != atom_mapping.end()) {
                      std::string mapped_name = mpos->second;
                      mapped_name[0] = 'H';
                      new_atom_mapping.insert(std::make_pair((*pos)->atmtype(), mapped_name));
                      continue;
                 }
            }
            extra_atom_set.insert((*pos)->atmtype());
            std::string warning = "Residue (" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + ") in Model " + String::IntToString(Mol_ID)
                                + ": Atom " + (*pos)->atmtype() + " is not mapped.\n";
            messages.push_back(warning);
            if ((*pos)->atom_type() != "H" && (*pos)->atom_type() != "D") _successful = false;
       }

       if (!extra_atom_set.empty()) {
            std::map<std::string, std::string> conflict_atom_mapping;
            conflict_atom_mapping.clear();

            for (std::map<std::string, std::string>::const_iterator pos = new_atom_mapping.begin(); pos != new_atom_mapping.end(); ++pos) {
                 // pos = atom_mapping.begin(); pos != atom_mapping.end(); ++pos) {
                 if (extra_atom_set.find(pos->second) == extra_atom_set.end()) {
                      // new_atom_mapping.insert(std::make_pair(pos->first, pos->second));
                      continue;
                 }
                 conflict_atom_mapping.insert(std::make_pair(pos->first, pos->second));
                 std::string error = "Residue (" + _pdb_chnid + " " + _ResName + " " + _pdb_res_no + _ins_code + ") in Model "
                                   + String::IntToString(Mol_ID) + ": mapping Atom " + pos->first + " to " + pos->second + " which already existed.\n";
                 messages.push_back(error);
            }

            if (!_successful) {
                 if (!messages.empty()) {
                      for (std::vector<std::string>::const_iterator
                           pos = messages.begin(); pos != messages.end(); ++pos) {
                           _logIo->message(pos->c_str());
                      }
                 }
                 return _successful;
            }

            if (!conflict_atom_mapping.empty()) {
                 for (std::map<std::string, std::string>::const_iterator pos = conflict_atom_mapping.begin(); pos != conflict_atom_mapping.end(); ++pos) {
                      if (_connected_to_same_atom(pos->first, pos->second)) {
                           new_atom_mapping.insert(std::make_pair(pos->second, pos->second));
                           extra_atom_set.erase(pos->second);
                           extra_atom_set.insert(pos->first);
                      } else new_atom_mapping.insert(std::make_pair(pos->first, pos->second));
                 }
            }

            std::set<std::string> used_atom_names;
            used_atom_names.clear();
            for (std::map<std::string, std::string>::const_iterator pos = new_atom_mapping.begin(); pos != new_atom_mapping.end(); ++pos) {
                 used_atom_names.insert(pos->second);
            }

            int count = 0;
            for (std::set<std::string>::const_iterator pos = extra_atom_set.begin(); pos != extra_atom_set.end(); ++pos) {
                 if (used_atom_names.find(*pos) == used_atom_names.end()) {
                      new_atom_mapping.insert(std::make_pair(*pos, *pos));
                      used_atom_names.insert(*pos);
                      continue;
                 }
                 std::string type = "H";
                 if ((*pos)[0] == 'D') type = "D";
                 while (true) {
                      count++;
                      std::string name = type + String::IntToString(count);
                      if (used_atom_names.find(name) == used_atom_names.end()) {
                           used_atom_names.insert(name);
                           new_atom_mapping.insert(std::make_pair(*pos, name));
                           break;
                      }
                 }
            }
       } // else new_atom_mapping = atom_mapping;

       if (!_successful) {
            if (!messages.empty()) {
                 for (std::vector<std::string>::const_iterator pos = messages.begin(); pos != messages.end(); ++pos) {
                      _logIo->message(pos->c_str());
                 }
            }
            return _successful;
       }

       for (std::map<std::string, std::string>::const_iterator pos = new_atom_mapping.begin(); pos != new_atom_mapping.end(); ++pos) {
            std::string cs = CompositeIndex::getIndex(_ResName, pos->first);
            std::pair<std::multimap<std::string, int>::iterator, std::multimap<std::string, int>::iterator> range = _atomNames.equal_range(cs);
            for (std::multimap<std::string, int>::iterator mpos = range.first; mpos != range.second; ++mpos) {
                 _atoms[mpos->second]->set_atmtype(pos->second);
                 _atoms[mpos->second]->set_pdb_atmnam(pos->second);
            }
       }
       if (index_flag) {
            _atomNameAlts_Orininal.clear();
            _UpdateIndices();
            reorder_atoms();
       }

       return _successful;
}

bool Residue::update_atom_mapping(const int& Mol_ID, const std::string& New_ResName, const std::map<std::string, std::string>& atom_mapping)
{
       if (!update_atom_name(atom_mapping, Mol_ID, false)) return false;

       if (_ResName != New_ResName) _OrigResName = _ResName;
       _ResName = New_ResName;

       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            (*pos)->set_restype(_ResName);
            (*pos)->set_pdb_resnam(_ResName);
       }

       if (_atomNameAlts_Orininal.empty()) _atomNameAlts_Orininal = _atomNameAlts;
       _UpdateIndices();
       reorder_atoms();

       return true;
}

void Residue::_createScratchGraph(const int& Mol_ID, std::vector<std::pair<std::string, std::string> >& atomlist,
                                  std::vector<std::pair<std::string, std::string> >& bondlist)
{
       atomlist.clear();
       bondlist.clear();

       std::map<std::string, std::string> name_type;
       _getAtomNameTypeMapping(name_type);

       std::string alt_loc = "";
       _getAtomList(alt_loc, name_type, atomlist);
       if (atomlist.size() < 2) {
            atomlist.clear();
            return;
       }

       _getBondList(alt_loc, name_type, bondlist);

       std::vector<std::set<std::string> > atom_sets;
       _checkGraphConnectivity(atomlist, bondlist, atom_sets);
       if (atom_sets.size() < 2) return;

       _findLinkBetweenSubGraph(alt_loc, name_type, atom_sets, atomlist, bondlist);
       if (atom_sets.size() < 2) return;

       atomlist.clear();
       bondlist.clear();

       std::string cs = "";
       for (std::vector<std::set<std::string> >::iterator pos = atom_sets.begin(); pos != atom_sets.end(); ++pos) {
            for (std::set<std::string>::iterator s1 = pos->begin(); s1 != pos->end(); ++s1) {
                 cs += *s1 + " ";
            }
            cs += "\n";
       }

       if (Mol_ID < 0)
            cs += "\nResidue (";
       else cs += "\nIn model " + String::IntToString(Mol_ID) + ", Residue (";
       cs += GetFirstAtom()->pdb_chnid() + " " + GetFirstAtom()->pdb_resnam() + " " + GetFirstAtom()->pdb_resnum() + ") is not connected.";
       _messageIo->insertMessage("residue_match_print", "warning", cs);
       cs += "\n";
       _logIo->message(cs.c_str());
}

void Residue::_getAtomNameTypeMapping(std::map<std::string, std::string>& name_type)
{
       name_type.clear();
       bool not_found_atom_type = false;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            name_type.insert(std::make_pair((*pos)->atmtype(), ""));
            if ((*pos)->atom_type().empty() || (*pos)->atom_type() == ".")
                 not_found_atom_type = true;
       }
       if (not_found_atom_type) {
            Element::getAtomSymbols(name_type);
       } else name_type.clear();
}

void Residue::_getAtomList(std::string& alt_loc, std::map<std::string, std::string>& name_type, std::vector<std::pair<std::string, std::string> >& atomlist)
{
       std::map<std::string, int> atomTree;
       atomTree.clear();
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            std::string type = (*pos)->atom_type();
            if (type.empty() || type == ".") {
                 std::map<std::string, std::string>::iterator mpos = name_type.find((*pos)->atmtype());
                 if (mpos != name_type.end()) type = mpos->second;
            }
            if (type == "Q") continue;

            if (!(*pos)->alt_loc().empty() && alt_loc.empty()) {
                 alt_loc = (*pos)->alt_loc();
            }
            if ((*pos)->alt_loc().empty() || (*pos)->alt_loc() == alt_loc) {
                 if (atomTree.find((*pos)->atmtype()) != atomTree.end()) continue;
                 atomTree.insert(std::make_pair((*pos)->atmtype(), atomlist.size()));
                 atomlist.push_back(std::make_pair(type, (*pos)->atmtype()));
            }
       }
}

void Residue::_getBondList(const std::string& alt_loc, std::map<std::string, std::string>& name_type,
                           std::vector<std::pair<std::string, std::string> >& bondlist)
{
       std::set<std::string> bondTree;
       bondTree.clear();
       for (unsigned int i = 0; i < _atoms.size() - 1; ++i) {
            if (!_atoms[i]->alt_loc().empty() && _atoms[i]->alt_loc() != alt_loc)
                 continue;

            std::string type_i = _atoms[i]->atom_type();
            if (type_i.empty() || type_i == ".") {
                 std::map<std::string, std::string>::iterator mpos = name_type.find(_atoms[i]->atmtype());
                 if (mpos != name_type.end()) type_i = mpos->second;
            }
            if (type_i == "Q") continue;
            for (unsigned int j = i + 1; j < _atoms.size(); ++j) {
                 if (!_atoms[j]->alt_loc().empty() && _atoms[j]->alt_loc() != alt_loc) continue;

                 std::string type_j = _atoms[j]->atom_type();
                 if (type_j.empty() || type_j == ".") {
                      std::map<std::string, std::string>::iterator mpos = name_type.find(_atoms[j]->atmtype());
                      if (mpos != name_type.end()) type_j = mpos->second;
                 }
                 if (type_j == "Q") continue;

                 if ((type_i == "H" || type_i == "D") && (type_j == "H" || type_j == "D")) continue;

                 if (!BondUtil::is_a_bond_with_extension(_atoms[i], type_i, _atoms[j], type_j)) continue;

                 std::string cs = _atoms[j]->atmtype() + "_" + _atoms[i]->atmtype();
                 if (_atoms[i]->atmtype() < _atoms[j]->atmtype()) cs = _atoms[i]->atmtype() + "_" + _atoms[j]->atmtype();
                 if (bondTree.find(cs) != bondTree.end()) continue;
                 bondTree.insert(cs);
                 bondlist.push_back(std::make_pair(_atoms[i]->atmtype(), _atoms[j]->atmtype()));
            }
       }
}

void Residue::_checkGraphConnectivity(std::vector<std::pair<std::string, std::string> >& atomlist,
                                      std::vector<std::pair<std::string, std::string> >& bondlist,
                                      std::vector<std::set<std::string> >& atom_sets)
{
       atom_sets.clear();
       std::set<std::string> atom_set;
       for (std::vector<std::pair<std::string, std::string> >::iterator pos = bondlist.begin(); pos != bondlist.end(); ++pos) {
            bool found = false;
            for (std::vector<std::set<std::string> >::iterator ptr = atom_sets.begin(); ptr != atom_sets.end(); ++ptr) {
                 if (ptr->find(pos->first) != ptr->end()) {
                      ptr->insert(pos->second);
                      found = true;
                      break;
                 } else if (ptr->find(pos->second) != ptr->end()) {
                      ptr->insert(pos->first);
                      found = true;
                      break;
                 }
            }
            if (!found) {
                 atom_set.clear();
                 atom_set.insert(pos->first);
                 atom_set.insert(pos->second);
                 atom_sets.push_back(atom_set);
            }
       }

       for (std::vector<std::pair<std::string, std::string> >::iterator pos = atomlist.begin(); pos != atomlist.end(); ++pos) {
            bool found = false;
            for (std::vector<std::set<std::string> >::iterator ptr = atom_sets.begin(); ptr != atom_sets.end(); ++ptr) {
                 if (ptr->find(pos->second) != ptr->end()) {
                      found = true;
                      break;
                 }
            }
            if (!found) {
                 atom_set.clear();
                 atom_set.insert(pos->second);
                 atom_sets.push_back(atom_set);
            }
       }

       if (atom_sets.size() <= 1) return;

       std::vector<std::set<std::string> > merged_atom_sets;
       merged_atom_sets.clear();
       while (true) {
            bool changed = false;
            for (unsigned int i = 0; i < atom_sets.size() - 1; ++i) {
                 if (atom_sets[i].empty()) continue;
                 for (unsigned int j = i + 1; j < atom_sets.size(); ++j) {
                      if (atom_sets[j].empty()) continue;
                      bool found = false;
                      for (std::set<std::string>::iterator pos = atom_sets[j].begin(); pos != atom_sets[j].end(); ++pos) {
                           if (atom_sets[i].find(*pos) != atom_sets[i].end()) {
                                found = true;
                                break;
                           }
                      }
                      if (found) {
                           for (std::set<std::string>::iterator pos = atom_sets[j].begin(); pos != atom_sets[j].end(); ++pos) {
                                atom_sets[i].insert(*pos);
                           }
                           atom_sets[j].clear();
                           changed = true;
                      }
                 }
            }
            if (!changed) break;
       }

       std::multimap<unsigned int, unsigned int> indexMap;
       indexMap.clear();
       for (unsigned int i = 0; i < atom_sets.size(); ++i) {
            if (atom_sets[i].empty()) continue;
            indexMap.insert(std::make_pair(atom_sets[i].size(), i));
       }

       for (std::multimap<unsigned int, unsigned int>::reverse_iterator rpos = indexMap.rbegin(); rpos != indexMap.rend(); ++rpos) {
            merged_atom_sets.push_back(atom_sets[rpos->second]);
       }
       atom_sets = merged_atom_sets;
}

void Residue::_findLinkBetweenSubGraph(const std::string& alt_loc, std::map<std::string, std::string>& name_type,
                        std::vector<std::set<std::string> >& atom_sets, std::vector<std::pair<std::string, std::string> >& atomlist,
                        std::vector<std::pair<std::string, std::string> >& bondlist)
{
       int maximum_time = atom_sets.size();
       for (int loop = 0; loop < maximum_time; ++loop) {
            Atom *found_atom1 = NULL;
            Atom *found_atom2 = NULL;
            std::string type_1 = "";
            std::string type_2 = "";
            double minimum_dist = 100;
            for (unsigned int k = 0; k < atom_sets.size() - 1; k++) {
                 for (std::set<std::string>::iterator s1 = atom_sets[k].begin(); s1 != atom_sets[k].end(); ++s1) {
                      Atom *atom1 = find_atom(*s1, alt_loc);
                      if (!atom1) continue;
                      
                      for (unsigned int l = k + 1; l < atom_sets.size(); l++) {
                           for (std::set<std::string>::iterator s2 = atom_sets[l].begin(); s2 != atom_sets[l].end(); ++s2) {
                                Atom *atom2 = find_atom(*s2, alt_loc);
                                if (!atom2) continue;

                                double dist = cal_distance(atom1, atom2);
                                if (dist < minimum_dist) {
                                     type_1 = atom1->atom_type();
                                     if (type_1.empty() || type_1 == ".") {
                                          std::map<std::string, std::string>::iterator mpos = name_type.find(atom1->atmtype());
                                          if (mpos != name_type.end()) type_1 = mpos->second;
                                     }
                                     type_2 = atom2->atom_type();
                                     if (type_2.empty() || type_2 == ".") {
                                          std::map<std::string, std::string>::iterator mpos = name_type.find(atom2->atmtype());
                                          if (mpos != name_type.end()) type_2 = mpos->second;
                                     }
                                     if (!type_1.empty() && type_1 != "." && !type_2.empty() && type_2 != ".") {
                                          minimum_dist = dist;
                                          found_atom1 = atom1;
                                          found_atom2 = atom2;
                                     }
                                }
                           }
                      }
                 }
            }
            if (!found_atom1 || !found_atom2) continue;

            if (!BondUtil::is_a_link_loose(type_1, type_2, minimum_dist)) continue;
               
            bondlist.push_back(std::make_pair(found_atom1->atmtype(), found_atom2->atmtype()));
            _checkGraphConnectivity(atomlist, bondlist, atom_sets);
            if (atom_sets.size() == 1) break;
       }
}

bool Residue::_connected_to_same_atom(const std::string& first_atom, const std::string& second_atom)
{
       Atom *fst = find_atom(first_atom);
       if (!fst) return false;
       Atom *fst_connected = _find_connected_atom(fst);
       if (!fst_connected) return false; 
       Atom *snd = find_atom(second_atom);
       if (!snd) return false;
       Atom *snd_connected = _find_connected_atom(snd);
       if (!snd_connected) return false;
       if (fst_connected == snd_connected) return true;
       return false;
}

Atom* Residue::_find_connected_atom(Atom* atom)
{
       Atom *found_atom = NULL;
       double distance = 100000.0;
       for (std::vector<Atom*>::iterator pos = _atoms.begin(); pos != _atoms.end(); ++pos) {
            if (*pos == atom || (*pos)->atom_type() == "H") continue;
            if (atom->alt_loc().empty() || (*pos)->alt_loc().empty() || atom->alt_loc() == (*pos)->alt_loc()) {
                 double dist = cal_distance(atom, *pos);
                 if (dist > 1.4) continue;
                 if (dist < distance) {
                      distance = dist;
                      found_atom = *pos;
                 }
            }
       }
       return found_atom;
}

bool is_connect(const std::string &atm_name_a, Residue *res_a, const std::string &atm_name_b, Residue *res_b)
{
       std::vector<Atom*> atom_a, atom_b;
       res_a->find_atom(atm_name_a, atom_a);
       res_b->find_atom(atm_name_b, atom_b);
       for (std::vector<Atom*>::iterator pos1 = atom_a.begin(); pos1 != atom_a.end(); ++pos1) {
            for (std::vector<Atom*>::iterator pos2 = atom_b.begin(); pos2 != atom_b.end(); ++pos2) {
                 if (!(*pos1)->alt_loc().empty() && !(*pos2)->alt_loc().empty() && (*pos1)->alt_loc() != (*pos2)->alt_loc()) continue;
                 if (BondUtil::is_a_bond(*pos1, *pos2)) return true;
            }
       }
       return false;
}

bool is_connect(const std::string &atm_name_a, Residue *res_a, const std::string &atm_name_b, Residue *res_b, const double cut_off)
{
       std::vector<Atom*> atom_a, atom_b;
       res_a->find_atom(atm_name_a, atom_a);
       res_b->find_atom(atm_name_b, atom_b);
       for (std::vector<Atom*>::iterator pos1 = atom_a.begin(); pos1 != atom_a.end(); ++pos1) {
            for (std::vector<Atom*>::iterator pos2 = atom_b.begin(); pos2 != atom_b.end(); ++pos2) {
                 if (!(*pos1)->alt_loc().empty() && !(*pos2)->alt_loc().empty() && (*pos1)->alt_loc() != (*pos2)->alt_loc()) continue;
                 double dist = cal_distance(*pos1, *pos2);
                 if (dist <= cut_off)  return true;
            }
       }
       return false;
}

bool is_connect(Residue *res_a, const std::string &atm_type_a, Residue *res_b, const std::string &atm_type_b)
{
       std::vector<Atom*> atom_a, atom_b;
       res_a->find_atom_by_type(atm_type_a, atom_a);
       res_b->find_atom_by_type(atm_type_b, atom_b);
       for (std::vector<Atom*>::iterator pos1 = atom_a.begin(); pos1 != atom_a.end(); ++pos1) {
            for (std::vector<Atom*>::iterator pos2 = atom_b.begin(); pos2 != atom_b.end(); ++pos2) {
                 if (!(*pos1)->alt_loc().empty() && !(*pos2)->alt_loc().empty() && (*pos1)->alt_loc() != (*pos2)->alt_loc()) continue;
                 if (BondUtil::is_a_bond(*pos1, *pos2)) return true;
            }
       }
       return false;
}

bool is_connect(Residue *res_a, const std::string &atm_type_a, Residue *res_b, const std::string &atm_type_b, std::vector<std::pair<Atom*, Atom*> >& pair_list)
{
       bool found = false;
       std::vector<Atom*> atom_a, atom_b;
       res_a->find_atom_by_type(atm_type_a, atom_a);
       res_b->find_atom_by_type(atm_type_b, atom_b);
       for (std::vector<Atom*>::iterator pos1 = atom_a.begin(); pos1 != atom_a.end(); ++pos1) {
            for (std::vector<Atom*>::iterator pos2 = atom_b.begin(); pos2 != atom_b.end(); ++pos2) {
                 if (!(*pos1)->alt_loc().empty() && !(*pos2)->alt_loc().empty() && (*pos1)->alt_loc() != (*pos2)->alt_loc()) continue;
                 if (BondUtil::is_a_bond(*pos1, *pos2)) {
                      pair_list.push_back(std::make_pair(*pos1, *pos2));
                      found = true;
                 }
            }
       }
       return found;
}

bool is_connect(Residue *res1, Residue *res2)
{
       for (std::vector<Atom*>::const_iterator pos1 = res1->_atoms.begin(); pos1 != res1->_atoms.end(); ++pos1) {
            for (std::vector<Atom*>::const_iterator pos2 = res2->_atoms.begin(); pos2 != res2->_atoms.end(); ++pos2) {
                 if (!(*pos1)->alt_loc().empty() && !(*pos2)->alt_loc().empty() && (*pos1)->alt_loc() != (*pos2)->alt_loc()) continue;
                 if (BondUtil::is_a_bond(*pos1, *pos2)) return true;
            }
       }
       return false;
}

bool is_polymer_connect(const std::string& chain_type, Residue *res1, Residue *res2)
{
       std::string F_atom = "C";
       std::string F_atom_alt = "";
       std::string S_atom = "N";
       double dist_cutoff = 1.76;
       if (chain_type == "ATOMN") {
            F_atom = "O3'";
            F_atom_alt = "O2'";
            S_atom = "P";
            dist_cutoff = 1.90;
       }

       if (is_connect(F_atom, res1, S_atom, res2, dist_cutoff)) return true;
       if (!F_atom_alt.empty() && is_connect(F_atom_alt, res1, S_atom, res2, dist_cutoff)) return true;

       return false;
}
